#!/usr/bin/env python
"""
_Workflow_

A simple object representing a Workflow in WMBS.

A workflow has an owner (e.g. PA instance, CRAB server user) and
a specification. The specification describes how jobs should be
created and what the jobs are supposed to do. This description
is held external to WMBS, WMBS just stores a pointer (url) to
the specification file. A workflow can be used with many
filesets and subscriptions (e.g. repeating the same task on a
bunch of data).

workflow + fileset = subscription
"""

import logging

from WMCore.DataStructs.Workflow import Workflow as WMWorkflow
from WMCore.WMBS.Fileset import Fileset
from WMCore.WMBS.WMBSBase import WMBSBase


class Workflow(WMBSBase, WMWorkflow):
    """
    A simple object representing a Workflow in WMBS.

    A workflow has an owner (e.g. PA instance, CRAB server user) and
    a specification. The specification describes how jobs should be
    created and what the jobs are supposed to do. This description
    is held external to WMBS, WMBS just stores a pointer (url) to
    the specification file. A workflow can be used with many
    filesets and subscriptions (e.g. repeating the same task on a
    bunch of data).

    workflow + fileset = subscription
    """

    def __init__(self, spec=None, owner="unknown", dn="unknown",
                 group="unknown", owner_vogroup="DEFAULT",
                 owner_vorole="DEFAULT", name=None, task=None,
                 wfType=None, id=-1, alternativeFilesetClose=False,
                 priority=None):
        WMBSBase.__init__(self)
        WMWorkflow.__init__(self, spec=spec, owner=owner, dn=dn,
                            group=group, owner_vogroup=owner_vogroup,
                            owner_vorole=owner_vorole, name=name,
                            task=task, wfType=wfType, priority=priority)

        if self.dn == "unknown":
            self.dn = owner

        self.id = id
        self.alternativeFilesetClose = alternativeFilesetClose
        return

    def exists(self):
        """
        _exists_

        Determine whether or not a workflow exists with the given spec, owner
        and name.  Return the ID if the workflow exists, False otherwise.
        """
        action = self.daofactory(classname="Workflow.Exists")
        result = action.execute(spec=self.spec, owner=self.dn,
                                group_name=self.vogroup,
                                role_name=self.vorole,
                                name=self.name, task=self.task,
                                conn=self.getDBConn(),
                                transaction=self.existingTransaction())

        return result

    def insertUser(self):
        """
        _insertUser_

        Check if the user exists in the wmbs database and return the id to be added
        in the workflow entry. If the user does not exists then it gets created.
        """
        existingTransaction = self.beginTransaction()

        userfactory = self.daofactory(classname="Users.GetUserId")
        userid = userfactory.execute(dn=self.dn,
                                     group_name=self.vogroup,
                                     role_name=self.vorole,
                                     conn=self.getDBConn(),
                                     transaction=self.existingTransaction())
        if not userid:
            newuser = self.daofactory(classname="Users.New")
            userid = newuser.execute(dn=self.dn, hn=self.owner,
                                     owner=self.owner, group=self.group,
                                     group_name=self.vogroup,
                                     role_name=self.vorole,
                                     conn=self.getDBConn(),
                                     transaction=self.existingTransaction())

        self.commitTransaction(existingTransaction)
        return userid

    def create(self):
        """
        _create_

        Write the workflow to the database.  If the workflow already exists in
        the database nothing will happen.
        """
        if self.exists() is not False:
            self.load()
            return

        userid = self.insertUser()

        existingTransaction = self.beginTransaction()
        action = self.daofactory(classname="Workflow.New")
        action.execute(spec=self.spec, owner=userid, name=self.name,
                       task=self.task, wfType=self.wfType,
                       alt_fs_close=self.alternativeFilesetClose,
                       priority=self.priority,
                       conn=self.getDBConn(),
                       transaction=self.existingTransaction())

        self.id = self.exists()
        self.commitTransaction(existingTransaction)
        logging.info("Workflow id %d created for %s", self.id, self.name)
        return

    def delete(self):
        """
        _delete_

        Remove this workflow from the database.
        """
        action = self.daofactory(classname="Workflow.Delete")
        action.execute(id=self.id, conn=self.getDBConn(),
                       transaction=self.existingTransaction())

        return

    def load(self):
        """
        _load_

        Load a workflow from WMBS.  One of the following must be provided:
          - The workflow ID
          - The workflow name and task
          - The workflow spec and owner and task
        """
        existingTransaction = self.beginTransaction()

        if self.id > 0:
            action = self.daofactory(classname="Workflow.LoadFromID")
            result = action.execute(workflow=self.id,
                                    conn=self.getDBConn(),
                                    transaction=self.existingTransaction())
        elif self.name is not None:
            action = self.daofactory(classname="Workflow.LoadFromNameAndTask")
            result = action.execute(workflow=self.name, task=self.task,
                                    conn=self.getDBConn(),
                                    transaction=self.existingTransaction())
        else:
            action = self.daofactory(classname="Workflow.LoadFromSpecOwner")
            result = action.execute(spec=self.spec, dn=self.dn,
                                    task=self.task, conn=self.getDBConn(),
                                    transaction=self.existingTransaction())

        self.id = result["id"]
        self.spec = result["spec"]
        self.name = result["name"]
        self.owner = result["owner"]
        self.dn = result["dn"]
        self.group = result["grp"]
        self.vorole = result["vogrp"]
        self.vogroup = result["vorole"]
        self.task = result["task"]
        self.wfType = result["type"]
        self.priority = result["priority"]

        action = self.daofactory(classname="Workflow.LoadOutput")
        results = action.execute(workflow=self.id, conn=self.getDBConn(),
                                 transaction=self.existingTransaction())

        self.outputMap = {}
        for outputID in results:
            for outputMap in results[outputID]:
                outputFileset = Fileset(id=outputMap["output_fileset"])
                if outputMap["merged_output_fileset"] is not None:
                    mergedOutputFileset = Fileset(id=outputMap["merged_output_fileset"])
                else:
                    mergedOutputFileset = None

                if outputID not in self.outputMap:
                    self.outputMap[outputID] = []

                self.outputMap[outputID].append({"output_fileset": outputFileset,
                                                 "merged_output_fileset": mergedOutputFileset})

        self.commitTransaction(existingTransaction)
        return

    def addOutput(self, outputIdentifier, outputFileset,
                  mergedOutputFileset=None):
        """
        _addOutput_

        Associate an output of this workflow with a particular fileset.
        """
        existingTransaction = self.beginTransaction()

        if self.id is False:
            self.create()

        if outputIdentifier not in self.outputMap:
            self.outputMap[outputIdentifier] = []

        self.outputMap[outputIdentifier].append({"output_fileset": outputFileset,
                                                 "merged_output_fileset": mergedOutputFileset})

        action = self.daofactory(classname="Workflow.InsertOutput")
        if mergedOutputFileset is not None:
            mergedFilesetID = mergedOutputFileset.id
        else:
            mergedFilesetID = None

        action.execute(workflowID=self.id, outputIdentifier=outputIdentifier,
                       filesetID=outputFileset.id, mergedFilesetID=mergedFilesetID,
                       conn=self.getDBConn(),
                       transaction=self.existingTransaction())

        self.commitTransaction(existingTransaction)
        return

    def countWorkflowsBySpec(self):
        """
        _countWorkflowsBySpec_

        Count workflows that share our spec
        """

        existingTransaction = self.beginTransaction()
        action = self.daofactory(classname="Workflow.CountWorkflowBySpec")
        result = action.execute(spec=self.spec)
        self.commitTransaction(existingTransaction)

        return result



from __future__ import print_function, division
import logging
from WMCore.DAOFactory import DAOFactory
from WMCore.Services.PyCondor.PyCondorAPI import getScheddParamValue


def availableScheddSlots(dbi, logger=logging, condorFraction=1):
    """
    check executing jobs and compare with condor limit.
    return the difference -- executing jobs - (condor limit * condorFraction)
    """
    action = DAOFactory(package='WMCore.WMBS',
                        logger=logger,
                        dbinterface=dbi)(classname="Jobs.GetCountByState")
    executingJobs = int(action.execute("executing"))

    maxScheddJobs = getScheddParamValue("MAX_JOBS_PER_OWNER")

    if maxScheddJobs is None:
        logger.warning("Failed to retrieve 'MAX_JOBS_PER_OWNER' from HTCondor")
        return 0

    freeSubmitSlots = int(int(maxScheddJobs) * condorFraction - executingJobs)
    return freeSubmitSlots


#!/usr/bin/env python
"""
_ListRunningJobs_

Retrieve the name, time of last state change, couch record and current state for
all jobs that are in the following states:
  none
  new
  created
  submitted
  executing
"""




from WMCore.Database.DBFormatter import DBFormatter

class ListRunningJobs(DBFormatter):
    sql = """SELECT wmbs_job.name AS job_name,
                    wmbs_job.state_time AS timestamp,
                    wmbs_job.couch_record AS couch_record,
                    wmbs_job_state.name AS state FROM wmbs_job
               INNER JOIN wmbs_job_state ON
                 wmbs_job.state = wmbs_job_state.id
             WHERE wmbs_job_state.name = 'none' OR
                   wmbs_job_state.name = 'new' OR
                   wmbs_job_state.name = 'created' OR
                   wmbs_job_state.name = 'submitted' OR
                   wmbs_job_state.name = 'executing'"""

    def execute(self, conn = None, transaction = False):
        result = self.dbi.processData(self.sql, conn = conn,
                                      transaction = transaction)
        return self.formatDict(result)



#!/usr/bin/env python
"""
WorkQueue splitting by block

"""
from __future__ import division

import os
from WMCore.WorkQueue.Policy.Start.StartPolicyInterface import StartPolicyInterface
from WMCore.WorkQueue.WorkQueueExceptions import WorkQueueWMSpecError, WorkQueueNoWorkError
from WMCore.DataStructs.Mask import Mask
from copy import copy
from math import ceil, floor

__all__ = []


class MonteCarlo(StartPolicyInterface):
    """Split elements into blocks"""
    def __init__(self, **args):
        StartPolicyInterface.__init__(self, **args)


    def split(self):
        """Apply policy to spec"""
        # if not specified take standard defaults
        self.args.setdefault('SliceType', 'NumberOfEvents')
        self.args.setdefault('SliceSize', 1000)  # events per job
        self.args.setdefault('SubSliceType', 'NumberOfEventsPerLumi')
        self.args.setdefault('SubSliceSize', self.args['SliceSize'])  # events per lumi
        self.args.setdefault('MaxJobsPerElement', 1000)  # jobs per WQE
        self.args.setdefault('MaxLumisPerElement', os.environ.get('MAX_LUMIS_PER_WQE'))
        self.args.setdefault('blowupFactor', 1.0)  # Estimate of additional jobs following tasks.
                                                  # Total WQE tasks will be Jobs*(1+blowupFactor)
        noInputUpdate = self.initialTask.getTrustSitelists().get('trustlists')
        noPileupUpdate = self.initialTask.getTrustSitelists().get('trustPUlists')

        if not self.mask:
            self.mask = Mask(FirstRun=1,
                             FirstLumi=self.initialTask.getFirstLumi(),
                             FirstEvent=self.initialTask.getFirstEvent(),
                             LastRun=1,
                             LastEvent=self.initialTask.getFirstEvent() +
                                             self.initialTask.totalEvents() - 1)
        mask = Mask(**self.mask)

        # First let's initialize some parameters
        lumis_per_job = ceil(self.args['SliceSize'] / self.args['SubSliceSize'])
        totalLumisPerElement = int(self.args['MaxJobsPerElement']) * lumis_per_job

        if self.args['MaxLumisPerElement'] and totalLumisPerElement > int(self.args['MaxLumisPerElement']):
            # If there are too many lumis in the WQ element. reduce the number of jobs per element
            self.args['MaxJobsPerElement'] = floor(int(self.args['MaxLumisPerElement']) / lumis_per_job)
            if self.args['MaxJobsPerElement'] == 0:
                raise WorkQueueWMSpecError(self.wmspec, """Too many lumis (%s) in a job:
                                           Change 'SliceSize' / 'SubSliceSize'""" % lumis_per_job)

        stepSize = int(self.args['SliceSize']) * int(self.args['MaxJobsPerElement'])
        total = mask['LastEvent'] - mask['FirstEvent'] + 1
        lastAllowedEvent = mask['LastEvent']
        eventsAccounted = 0

        while eventsAccounted < total:
            current = mask['FirstEvent'] + stepSize - 1  # inclusive range
            if current > lastAllowedEvent:
                current = lastAllowedEvent
            mask['LastEvent'] = current

            # Calculate the job splitting without actually doing it
            # number of lumis is calculated by events number and SubSliceSize which is events per lumi
            # So if there no exact division between events per job and events per lumi
            # it takes the ceiling of the value.
            # Therefore total lumis can't be calculated from total events / SubSliceSize
            # It has to be caluated by adding the lumis_per_job * number of jobs
            nEvents = mask['LastEvent'] - mask['FirstEvent'] + 1
            nLumis = floor(nEvents / self.args['SliceSize']) * lumis_per_job
            remainingLumis = ceil(nEvents % self.args['SliceSize'] / self.args['SubSliceSize'])
            nLumis += remainingLumis
            jobs = ceil(nEvents / self.args['SliceSize'])

            mask['LastLumi'] = mask['FirstLumi'] + int(nLumis) - 1  # inclusive range
            self.newQueueElement(WMSpec=self.wmspec,
                                 NumberOfLumis=nLumis,
                                 NumberOfEvents=nEvents,
                                 Jobs=jobs,
                                 Mask=copy(mask),
                                 NoInputUpdate=noInputUpdate,
                                 NoPileupUpdate=noPileupUpdate,
                                 blowupFactor=self.args['blowupFactor'])


            if mask['LastEvent'] > (2 ** 32 - 1):
                # This is getting tricky, to ensure consecutive
                # events numbers we must calculate where the jobSplitter
                # will restart the firstEvent to 1 for the last time
                # in the newly created unit
                internalEvents = mask['FirstEvent']
                accumulatedEvents = internalEvents
                breakPoint = internalEvents

                while accumulatedEvents < mask['LastEvent']:
                    if (internalEvents + self.args['SliceSize'] - 1) > (2 ** 32 - 1):
                        internalEvents = 1
                        breakPoint = accumulatedEvents
                    else:
                        internalEvents += self.args['SliceSize']
                        accumulatedEvents += self.args['SliceSize']

                leftoverEvents = mask['LastEvent'] - breakPoint + 1
                mask['FirstEvent'] = leftoverEvents + 1

            else:
                mask['FirstEvent'] = mask['LastEvent'] + 1

            mask['FirstLumi'] = mask['LastLumi'] + 1
            eventsAccounted += stepSize
            lastAllowedEvent = (total - eventsAccounted) + mask['FirstEvent'] - 1


    def validate(self):
        """Check args and spec work with block splitting"""
        StartPolicyInterface.validateCommon(self)

        if self.initialTask.totalEvents() < 1:
            raise WorkQueueNoWorkError(self.wmspec, 'Invalid total events selection: %s' % str(self.initialTask.totalEvents()))

        if self.mask and self.mask['LastEvent'] < self.mask['FirstEvent']:
            raise WorkQueueWMSpecError(self.wmspec, "Invalid start & end events")

        if self.mask and self.mask['LastLumi'] < self.mask['FirstLumi']:
            raise WorkQueueWMSpecError(self.wmspec, "Invalid start & end lumis")



#!/usr/bin/env python
"""
_ResourceControl_

Library from manipulating and querying the resource control database.
"""

from builtins import str, bytes

import time
from WMCore.BossAir.BossAirAPI import BossAirAPI
from WMCore.DAOFactory import DAOFactory
from WMCore.WMConnectionBase import WMConnectionBase
from WMCore.WMException import WMException


class ResourceControlException(WMException):
    """
    _ResourceControlException_

    Exception for the ResourceControl mechanisms
    """
    pass


class ResourceControl(WMConnectionBase):
    def __init__(self, config=None):
        WMConnectionBase.__init__(self, daoPackage="WMCore.ResourceControl")
        self.wmbsDAOFactory = DAOFactory(package="WMCore.WMBS",
                                         logger=self.logger,
                                         dbinterface=self.dbi)
        self.config = config
        return

    def insertSite(self, siteName, pendingSlots=0, runningSlots=0,
                   pnn=None, ceName=None, cmsName=None,
                   plugin=None):
        """
        _insertSite_

        Insert a site into WMBS tables, in the following order:
         1. first add PSN/site info into wmbs_location
         2. then add PNNs into wmbs_pnns
         3. at last, create a mapping of PSN/PNN in wmbs_location_pnns
        """
        timeNow = int(time.time())
        insertAction = self.wmbsDAOFactory(classname="Locations.New")
        insertAction.execute(siteName=siteName, pendingSlots=pendingSlots,
                             runningSlots=runningSlots,
                             pnn=pnn, ceName=ceName,
                             plugin=plugin, cmsName=cmsName, stateTime=timeNow,
                             conn=self.getDBConn(),
                             transaction=self.existingTransaction())
        return

    def insertPNNs(self, pnns):
        """
        _insertPNNs_

        Insert a list of standalone PNNs into WMBS (usually used for MSS nodes)
        """
        addAction = self.wmbsDAOFactory(classname="Locations.AddPNNs")
        addAction.execute(pnns=pnns, conn=self.getDBConn(),
                          transaction=self.existingTransaction())
        return

    def changeSiteState(self, siteName, state):
        """
        _changeSiteState_
        Set a site to some of the possible states and perform
        proper actions with the jobs, according to the state
        """
        timeNow = int(time.time())
        state2ExitCode = {"Aborted": 71301,
                          "Draining": 71302,
                          "Down": 71303}
        executingJobs = self.wmbsDAOFactory(classname="Jobs.ListByState")
        jobInfo = executingJobs.execute(state='executing')

        if jobInfo:
            bossAir = BossAirAPI(self.config)
            jobtokill = bossAir.updateSiteInformation(jobInfo, siteName, state in state2ExitCode)

            ercode = state2ExitCode.get(state, 71300)
            bossAir.kill(jobtokill, errorCode=ercode)

        # only now that jobs were updated by the plugin, we flip the site state
        setStateAction = self.wmbsDAOFactory(classname="Locations.SetState")
        setStateAction.execute(siteName=siteName, state=state, stateTime=timeNow,
                               conn=self.getDBConn(),
                               transaction=self.existingTransaction())

        return

    def listCurrentSites(self):
        """
        _listCurrentSites_

        List all the sites currently in Resource Control
        """
        listAction = self.daofactory(classname="ListCurrentSites")
        sitesArray = listAction.execute(conn=self.getDBConn(),
                                        transaction=self.existingTransaction())
        return [entry['site'] for entry in sitesArray]

    def listSitesSlots(self):
        """
        _listSitesSlots_

        List all sites, their slots and state available in Resource Control
        """
        listAction = self.daofactory(classname="ListSitesSlotsState")
        results = listAction.execute(conn=self.getDBConn(),
                                     transaction=self.existingTransaction())
        return results

    def listSiteInfo(self, siteName=None):
        """
        List the site name, SE name, CE name, pending and running slots,
        plugin, cms name and state for a given site.
        :param siteName: a string with the site name
        :return: a list of dictionaries with site information
        """
        listAction = self.wmbsDAOFactory(classname="Locations.GetSiteInfo")
        results = listAction.execute(siteName=siteName,
                                     conn=self.getDBConn(),
                                     transaction=self.existingTransaction())
        return results

    def changeTaskPriority(self, taskType, priority):
        """
        _changeTaskPriority_

        Change the priority of a sub task in WMBS. If the task
        type doesn't exist already then it will be added first.
        """
        existingTransaction = self.beginTransaction()
        subTypeAction = self.wmbsDAOFactory(classname="Subscriptions.InsertType")
        subTypeAction.execute(subType=taskType,
                              priority=priority,
                              conn=self.getDBConn(),
                              transaction=existingTransaction)
        self.commitTransaction(existingTransaction)
        return

    def insertThreshold(self, siteName, taskType, maxSlots, pendingSlots):
        """
        _insertThreshold_

        Insert a threshold into the Resource Control database.  If the threshold
        already exists it will be updated.
        taskType may be a list of tasks. Update each task individually.
        """
        existingTransaction = self.beginTransaction()

        subTypeAction = self.wmbsDAOFactory(classname="Subscriptions.InsertType")
        insertAction = self.daofactory(classname="InsertThreshold")
        if isinstance(taskType, (str, bytes)):
            taskType = [taskType]
        for singleTask in taskType:
            subTypeAction.execute(subType=singleTask,
                                  conn=self.getDBConn(),
                                  transaction=self.existingTransaction())
            insertAction.execute(siteName=siteName,
                                 taskType=singleTask,
                                 maxSlots=maxSlots,
                                 pendingSlots=pendingSlots,
                                 conn=self.getDBConn(),
                                 transaction=self.existingTransaction())

        self.commitTransaction(existingTransaction)
        return

    def listThresholdsForSubmit(self):
        """
        _listThresholdsForSubmit_

        Retrieve a list of job threshold information as well as information on
        the number of jobs running for all the known sites.  This information is
        returned in the form of a two level dictionary.  The first level is
        keyed by the site name.  The second level has the following keys:
          cms_name            - CMS name of the site
          pnns                - List with associated PNNs
          state               - State of the site
          state_time          - Timestamp in which the site joined the current state
          total_pending_slots - Total number of pending slots available at the site
          total_running_slots - Total number of running slots available at the site
          total_pending_jobs  - Total jobs pending at the site
          total_running_jobs  - Total jobs running at the site
          thresholds          - List of dictionaries with threshold information ordered by descending priority
        The threshold dictionaries have the following keys:
          task_type           - Type of the task associated with the thresholds
          max_slots           - Maximum running slots for the task type
          pending_slots       - Maximum pending slots for the task type
          task_running_jobs   - Running jobs for the task type
          task_pending_jobs   - Pending jobs for the task type
          priority            - Priority assigned to the task type
          wf_highest_priority - highest priority value among pending and running joba
        """
        listAction = self.daofactory(classname="ListThresholdsForSubmit")
        return listAction.execute(conn=self.getDBConn(),
                                  transaction=self.existingTransaction())

    def listThresholdsForCreate(self):
        """
        _listThresholdsForCreate_

        This will return a three level dictionary with the first level being
        keyed by site name.  The second level will have the following keys:
          total_slots - Total number of pending slots available at the site
          pending_jobs - Total number of jobs pending at the site per priority level, it is a dictionary
        """
        listAction = self.daofactory(classname="ListThresholdsForCreate")
        return listAction.execute(conn=self.getDBConn(),
                                  transaction=self.existingTransaction())

    def listWorkloadsForTaskSite(self, taskType, siteName):
        """
        _listWorkflowsForTaskSite_

        For the given task and site list the number of jobs running for each
        task.
        """
        listActions = self.daofactory(classname="ListWorkloadsForTaskSite")
        return listActions.execute(taskType, siteName,
                                   conn=self.getDBConn(),
                                   transaction=self.existingTransaction())

    def setJobSlotsForSite(self, siteName, pendingJobSlots=None,
                           runningJobSlots=None):
        """
        _setJobSlotsForSite_

        Set the number of running and/or pending job slots for the given site.
        """
        if pendingJobSlots is not None:
            pendingSlotsAction = self.daofactory(classname="SetPendingJobSlotsForSite")
            pendingSlotsAction.execute(siteName, pendingJobSlots,
                                       conn=self.getDBConn(),
                                       transaction=self.existingTransaction())

        if runningJobSlots is not None:
            runningSlotsAction = self.daofactory(classname="SetRunningJobSlotsForSite")
            runningSlotsAction.execute(siteName, runningJobSlots,
                                       conn=self.getDBConn(),
                                       transaction=self.existingTransaction())

    def thresholdBySite(self, siteName):
        """
        _thresholdBySite_

        List the thresholds of a single site
        """
        listActions = self.daofactory(classname="ThresholdBySite")
        return listActions.execute(site=siteName,
                                   conn=self.getDBConn(),
                                   transaction=self.existingTransaction())



#!/usr/bin/env python
"""
_ListThresholds_

Query the database to determine the threshold for each site.
"""




from WMCore.Database.DBFormatter import DBFormatter

class ListThresholds(DBFormatter):
    sql = """SELECT wl.site_name AS site, wst.name AS type,
                    rc.max_slots, rc.pending_slots FROM rc_threshold rc
                INNER JOIN wmbs_location wl ON wl.id = rc.site_id
                INNER JOIN wmbs_sub_types wst ON wst.id = rc.sub_type_id
                ORDER By site
           """

    def execute(self, conn = None, transaction = False):
        results = self.dbi.processData(self.sql,
                                       conn = conn, transaction = transaction)
        return self.formatDict(results)



#!/usr/bin/env python

"""
Oracle implementation of AddChecksumByLFN
"""





from WMComponent.DBS3Buffer.MySQL.DBSBufferFiles.AddChecksumByLFN import AddChecksumByLFN as MySQLAddChecksumByLFN

class AddChecksumByLFN(MySQLAddChecksumByLFN):
    """

    Add Checksums using lfn as key

    """

    sql = """INSERT INTO dbsbuffer_file_checksums (fileid, typeid, cksum)
             SELECT (SELECT id FROM dbsbuffer_file WHERE lfn = :lfn),
             (SELECT id FROM dbsbuffer_checksum_type WHERE type = :cktype), :cksum FROM dual
             WHERE NOT EXISTS (SELECT fileid FROM dbsbuffer_file_checksums WHERE
                               fileid = (SELECT id FROM dbsbuffer_file WHERE lfn = :lfn)
                               AND typeid = (SELECT id FROM dbsbuffer_checksum_type WHERE type = :cktype))"""



#!/usr/bin/env python
"""
_MarkInjectedWorkflows_

MySQL implementation of Workflow.MarkInjectedWorkflows

"""
__all__ = []

from WMCore.Database.DBFormatter import DBFormatter

class MarkInjectedWorkflows(DBFormatter):
    """
    Marks workflows that have been fully injected into WMBS
    """
    selectSQL = """SELECT spec FROM wmbs_workflow WHERE name = :name"""
    updateSQL = """UPDATE wmbs_workflow SET injected = :injected WHERE spec = :spec"""


    def execute(self, names, injected = True, conn = None,
                transaction = False):
        """
        Update the workflows to match their injected status

        """

        if injected:
            injected = 1
        else:
            injected = 0

        # Get the specs
        binds = []
        for name in names:
            binds.append({'name': name})

        result = self.dbi.processData(self.selectSQL, binds, conn = conn,
                                      transaction = transaction)

        # Use specs to update table
        l = self.formatDict(result)
        binds = []
        for entry in l:
            binds.append({'spec': entry['spec'], 'injected': injected})


        self.dbi.processData(self.updateSQL, binds, conn = conn,
                             transaction = transaction)

        return



#!/usr/bin/env python
"""
_GetLocation_

MySQL implementation of File.GetLocation
"""




from WMCore.Database.DBFormatter import DBFormatter

class GetLocation(DBFormatter):
    sql = """select pnn from dbsbuffer_location
                where id in (select location from dbsbuffer_file_location
                    where filename in (select id from dbsbuffer_file where lfn=:lfn))"""


    def getBinds(self, file=None):
        binds = []
        file = self.dbi.makelist(file)
        for f in file:
            binds.append({'lfn': f})
        return binds

    def format(self, result):
        "Return a list of SE FQDN's"
        out = set()
        for r in result:
            for i in r.fetchall():
                out.add(i[0])
        return out

    def execute(self, file=None, conn = None, transaction = False):
        binds = self.getBinds(file)

        result = self.dbi.processData(self.sql, binds,
                         conn = conn, transaction = transaction)
        return self.format(result)



#!/usr/bin/env python
"""
_TaskSpace_

Frontend module for setting up TaskSpace & StepSpace areas within a job.
"""

import inspect
import logging
import os
import os.path
import pickle
import socket
import sys
import threading
import json
from logging.handlers import RotatingFileHandler

import WMCore.FwkJobReport.Report as Report
from WMCore.DataStructs.JobPackage import JobPackage
from WMCore.Storage.SiteLocalConfig import loadSiteLocalConfig, SiteConfigError
from WMCore.WMException import WMException
from WMCore.WMRuntime import StepSpace
from WMCore.WMRuntime import TaskSpace
from WMCore.WMRuntime.Watchdog import Watchdog
from WMCore.WMSpec.WMWorkload import WMWorkloadHelper
from WMCore.WMRuntime.Tools.Scram import getPlatformMachine
from WMCore.BossAir.Plugins.BasePlugin import BasePlugin

class BootstrapException(WMException):
    """ An awesome exception """
    pass


def readFloatFromFile(filePath):
    try:
        with open(filePath, "r") as nf:
            return float(nf.readline())
    except:
        return None


# Copied direct from ProdAgent to find the damn CE name
def getSyncCE():
    """
    _getSyncCE_

    Extract the SyncCE from GLOBUS_GRAM_JOB_CONTACT if available for OSG,
    otherwise broker info for LCG

    """
    result = socket.gethostname()

    if 'GLOBUS_GRAM_JOB_CONTACT' in os.environ:
        #  //
        # // OSG, Sync CE from Globus ID
        # //
        val = os.environ['GLOBUS_GRAM_JOB_CONTACT']
        try:
            host = val.split("https://", 1)[1]
            host = host.split(":")[0]
            result = host
        except Exception:
            logging.warning("Failed to extract SyncCE from globus")
        return result

    if 'NORDUGRID_CE' in os.environ:
        #  //
        # // ARC, Sync CE from env. var. submitted with the job by JobSubmitter
        # //
        return os.environ['NORDUGRID_CE']

    return result


def establishTaskSpace(**args):
    """
    _establishTaskSpace_

    Bootstrap method for the execution dir for a WMTask

    """
    return TaskSpace.TaskSpace(**args)


def establishStepSpace(**args):
    """
    _establishStepSpace_

    Bootstrap method for the execution dir of a WMStep within a WMTask

    """
    return StepSpace.StepSpace(**args)


def locateWMSandbox():
    """
    _locateWMSandbox_

    At runtime, the WMSandbox module should be defined and available
    on the PYTHONPATH.

    Look up the location of the module via import so that the
    pickled files in the sandbox can be loaded

    """
    try:
        import WMSandbox
    except ImportError as ex:
        msg = "Error importing WMSandbox module"
        msg += str(ex)
        raise BootstrapException(msg)

    wmsandboxLoc = inspect.getsourcefile(WMSandbox)
    wmsandboxLoc = wmsandboxLoc.replace("__init__.py", "")
    return wmsandboxLoc


def loadJobDefinition():
    """
    _loadJobDefinition_

    Load the job package and pull out the indexed job, return
    WMBS Job instance

    Although this will create a JobReport, it won't necessarily bring it back.
    Report names are dependent on the retry_count, but if it fails unpacking the job
    it doesn't know the retry_count and will create the wrong file
    """
    sandboxLoc = locateWMSandbox()
    package = JobPackage()
    packageLoc = os.path.join(sandboxLoc, "JobPackage.pcl")
    try:
        package.load(packageLoc)
    except Exception as ex:
        msg = "Failed to load JobPackage:%s\n" % packageLoc
        msg += str(ex)
        createErrorReport(exitCode=11001, errorType="JobPackageError", errorDetails=msg)
        raise BootstrapException(msg)

    try:
        import WMSandbox.JobIndex
    except ImportError as ex:
        msg = "Failed to import WMSandbox.JobIndex module\n"
        msg += str(ex)
        createErrorReport(exitCode=11002, errorType="JobIndexError", errorDetails=msg)
        raise BootstrapException(msg)

    index = WMSandbox.JobIndex.jobIndex

    try:
        job = package[index]
    except Exception:
        msg = "Failed to extract job index %i " % index
        msg += "from the jobPackage directory: %s\n" % package.get('directory')
        msg += "Found a total of %d indexes in the JobPackage.\n" % len(package)
        createErrorReport(exitCode=11003, errorType="JobExtractionError", errorDetails=msg)
        raise BootstrapException(msg)
    logging.info("Job Index = %s\nJob Instance = %s\n", index, job)

    return job


def loadWorkload():
    """
    _loadWorkload_

    Load the Workload from the WMSandbox Area

    """
    sandboxLoc = locateWMSandbox()
    workloadPcl = "%s/WMWorkload.pkl" % sandboxLoc
    with open(workloadPcl, 'rb') as handle:
        wmWorkload = pickle.load(handle)

    return WMWorkloadHelper(wmWorkload)


def loadTask(job):
    """
    _loadTask_

    load the Workload, and then lookup the task in the workload
    required by the job

    """
    workload = loadWorkload()

    try:
        task = workload.getTaskByPath(job['task'])
    except KeyError as ex:
        msg = "Task name not in job object"
        msg += str(ex)
        createErrorReport(exitCode=11103, errorType="TaskNotInJob", errorDetails=msg,
                          logLocation="Report.%i.pkl" % job['retry_count'])
        raise BootstrapException(msg)
    except Exception as ex:
        msg = "Error looking up task %s\n" % job['task']
        msg += str(ex)
        createErrorReport(exitCode=11101, errorType="TaskLookupError", errorDetails=msg,
                          logLocation="Report.%i.pkl" % job['retry_count'])
        raise BootstrapException(msg)
    if task is None:
        msg = "Unable to look up task %s from Workload\n" % job['task']
        msg += "Task name not matched"
        createErrorReport(exitCode=11102, errorType="TaskNotFound", errorDetails=msg,
                          logLocation="Report.%i.pkl" % job['retry_count'])
        raise BootstrapException(msg)
    return task

def createWMRuntimeJson(outputPath, jobTypes=("Production", "Processing")):
    """
    Create a json with runtime information in the following form below:
    {
     "workflow_type: "Stepchain",  # string with the request type
     "number_of_cmsRuns": 2,  # an integer >= 0
     "worker_arch": "X86_64",  # string with the architecture of the worker node
     "worker_os": "rhel9",  # string with the OS of the worker node
     "job_type": "Production", # string with job type information: Production/Processing, Merge, LogCollect, etc.
     "cmsRun_params": [{
         "step": "cmsRun1",  # string with the relevant step
         "keek_output": boolean saying whether we keep output or not,
         "input_files": [string of input LFNs or a single local file],
         "output_files": [ordered string of output local files],
         "output_files_datatiers": [ordered string of output local files datatiers],
         "transient_output": [ordered boolean saying whether output files are announced in the workflow or not],
         "output_files_unmerged_base": [ordered string with lfnBase],
         "output_files_unmerged_base": [ordered string with mergedLFNBase info]
         },
         {
         "step": "cmsRun2",  # string with the relevant step
         # and so on...
         },
         # any new step appends step data to this list
     ]
     } 
     :param outputPath: path to write the json
     :param jobTypes: tuple of job types to send to logging. E.g.: ("Production", "Processing", "Merge", "LogCollect")
     :return
    """
    # Load workflow-level information
    workload = loadWorkload()
    job = loadJobDefinition()
    task = loadTask(job)
    # Create dictionary to collect runtime info
    runtimeInfo = {}
    runtimeInfo['workflow_type'] = workload.getRequestType()
    runtimeInfo['number_of_cmsRuns'] = len(task.listAllStepNames(cmsRunOnly=True))
    runtimeInfo['worker_arch'] = getPlatformMachine()
    runtimeInfo['worker_os'] = BasePlugin.scramArchtoRequiredOS(task.getScramArch())
    runtimeInfo['job_type'] = job.get('jobType', None)
    cmsRun_params = []
    # Jobs with MCFakeFile means no input files
    mcFakeFile=False
    if len(job['input_files']) == 1:
        if job['input_files'][0]['lfn'].startswith("MCFakeFile"):
            mcFakeFile=True
    # Get information per cmsRun step
    for cmsswStep in task.listAllStepNames(cmsRunOnly=True):
        cmsRunParam = {}
        step = task.getStepHelper(cmsswStep)
        output_modules = step.listOutputModules()
        keepOutput = getattr(step.data.output, 'keep')
        # Collect input files
        inputFileNames = []
        if mcFakeFile:
            # For chained processes, point to previous step output
            chained = getattr(step.data.input, 'chainedProcessing', False)
            if chained:
                inputModule = getattr(step.data.input, 'inputOutputModule', None)
                inputStepName = getattr(step.data.input, 'inputStepName', None)
                inputFileNames.append("../{}/{}.root".format(inputStepName, inputModule))
        else:
            # Get input files from job package information
            for inputFiles in job.get('input_files', []):
                inputFileNames.append(inputFiles['lfn'])

        # Collect output files per module
        outputFileNames = []
        outputFilesDataTiers = []
        outputFilesTransient = []
        outputFilesLFNBase = []
        outputFilesMergedLFNBase = []
        for moduleName in output_modules:
            outputModule = step.getOutputModule(moduleName)
            output_filename = "{}.root".format(moduleName)
            outputFileNames.append(output_filename)
            outputFilesDataTiers.append(getattr(outputModule, 'dataTier', None))
            outputFilesTransient.append(getattr(outputModule, 'transient', None))
            outputFilesLFNBase.append(getattr(outputModule, 'lfnBase', None))
            outputFilesMergedLFNBase.append(getattr(outputModule, 'mergedLFNBase', None))

        cmsRunParam['step'] = step.stepName()
        cmsRunParam['keep_output'] = keepOutput
        cmsRunParam['input_files'] = inputFileNames
        cmsRunParam['output_files'] = outputFileNames
        cmsRunParam['output_files_datatiers'] = outputFilesDataTiers
        cmsRunParam['transient_output'] = outputFilesTransient
        cmsRunParam['output_files_unmerged_base'] = outputFilesLFNBase
        cmsRunParam['output_files_merged_base'] = outputFilesMergedLFNBase
        cmsRun_params.append(cmsRunParam)
    runtimeInfo['cmsRun_params'] = cmsRun_params

    jsonPath = "{}/runtimeInfo.json".format(outputPath)
    with open(jsonPath, "w", encoding="utf-8") as f:
        json.dump(runtimeInfo, f, indent=4, sort_keys=True)

    # Log Production/Processing
    if runtimeInfo['job_type'] in jobTypes:
        logging.info('runtime json = {}'.format(json.dumps(runtimeInfo, indent=4, sort_keys=True)))

    return

def createInitialReport(job, reportName):
    """
    _createInitialReport_

    Create an initial job report with the base
    information in it.
    """
    try:
        siteCfg = loadSiteLocalConfig()
    except SiteConfigError:
        # For now, assume that we did this on purpose
        msg = "Couldn't find SiteConfig"
        logging.error(msg)
        # TODO: Make less goatballs for testing purposes
        return

    report = Report.Report()

    report.data.WMAgentJobID = job.get('id', None)
    report.data.WMAgentJobName = job.get('name', None)
    report.data.pnn = siteCfg.localStageOut.get('phedex-node', 'Unknown')
    report.data.siteName = getattr(siteCfg, 'siteName', 'Unknown')
    report.data.hostName = socket.gethostname()
    report.data.ceName = getSyncCE()

    # TODO: need to check what format it returns and what features need to extract.
    # currently
    # $MACHINEFEATURES/hs06: HS06 score of the host
    # $MACHINEFEATURES/total_cpu: number of configured job slots
    # $JOBFEATURES/hs06_job: HS06 score available to your job
    # $JOBFEATURES/allocated_cpu: number of allocated slots (=8 in case of a multicore job

    machineFeaturesFile = os.environ.get('MACHINEFEATURES')
    report.data.machineFeatures = {}
    if machineFeaturesFile:
        report.data.machineFeatures['hs06'] = readFloatFromFile("%s/hs06" % machineFeaturesFile)
        report.data.machineFeatures['total_cpu'] = readFloatFromFile("%s/total_cpu" % machineFeaturesFile)

    jobFeaturesFile = os.environ.get('JOBFEATURES')
    report.data.jobFeatures = {}
    if jobFeaturesFile:
        report.data.jobFeatures['hs06_job'] = readFloatFromFile("%s/hs06_job" % jobFeaturesFile)
        report.data.jobFeatures['allocated_cpu'] = readFloatFromFile("%s/allocated_cpu" % jobFeaturesFile)

    report.data.completed = False
    report.setTaskName(taskName=job.get('task', 'TaskNotFound'))

    # Not so fond of this, but we have to put the master
    # report way up at the top so it's returned if the
    # job fails early
    reportPath = os.path.join(os.getcwd(), '../', reportName)
    report.save(reportPath)

    return


def createErrorReport(exitCode, errorType, errorDetails=None,
                      logLocation="Report.0.pkl"):
    """
    _createErrorReport_

    Create a report if something fails inside the Bootstrap
    This creates a dummy step called 'CRITICAL' and
    sticks the error in there.
    """

    try:
        siteCfg = loadSiteLocalConfig()
    except SiteConfigError:
        # For now, assume that we did this on purpose
        msg = "Couldn't find SiteConfig"
        logging.error(msg)
        # TODO: Make this not suck goatballs when you are just running tests
        return
    report = Report.Report()

    report.data.pnn = siteCfg.localStageOut.get('phedex-node', 'Unknown')
    report.data.siteName = getattr(siteCfg, 'siteName', 'Unknown')
    report.data.hostName = socket.gethostname()
    report.data.ceName = getSyncCE()
    report.data.completed = False

    report.addError(stepName='CRITICAL', exitCode=exitCode,
                    errorType=errorType, errorDetails=errorDetails)

    reportPath = os.path.join(os.getcwd(), '../', logLocation)
    report.save(reportPath)

    return


def setupLogging(logDir, logName="wmagentJob.log", useStdout=False):
    """
    _setupLogging_

    Setup logging for the slave process. Each slave process will have its own
    log file.
    useStdout adds a stdout handler to the logger object such that records are
    also written to the main job log object.
    """
    try:
        # create a root logger
        logger = logging.getLogger()
        # create a log formatter
        logFormatter = logging.Formatter("%(asctime)s:%(levelname)s:%(module)s:%(message)s")

        if useStdout:
            consoleHandler = logging.StreamHandler(sys.stdout)
            consoleHandler.setFormatter(logFormatter)
            logger.addHandler(consoleHandler)
        else:
            # create a file handler
            logFile = "%s/%s" % (logDir, logName)
            logHandler = RotatingFileHandler(logFile, "a", 1000000000, 3)
            logHandler.setFormatter(logFormatter)
            logger.addHandler(logHandler)

        logger.setLevel(logging.INFO)

        # This is left in as a reminder for debugging purposes
        # SQLDEBUG turns your log files into horrible messes
        # logging.getLogger().setLevel(logging.SQLDEBUG)
        myThread = threading.currentThread()
        myThread.logger = logger
    except Exception as ex:
        msg = "Error setting up logging in dir %s:\n" % logDir
        msg += str(ex)
        raise BootstrapException(msg)
    return


def setupMonitoring(logName):
    """
    Setup the basics of the watchdog monitoring.
    Attach it to a thread.

    """
    try:
        monitor = Watchdog(logPath=logName)
        myThread = threading.currentThread
        myThread.watchdogMonitor = monitor
        return monitor
    except Exception as ex:
        msg = "Error setting up Watchdog monitoring:\n"
        msg += str(ex)
        raise BootstrapException(msg)



#!/usr/bin/env python

"""
MySQL implementation of AddChecksumByLFN
"""

from WMCore.Database.DBFormatter import DBFormatter


class AddChecksumByLFN(DBFormatter):
    sql = """INSERT IGNORE INTO dbsbuffer_file_checksums (fileid, typeid, cksum)
             SELECT (SELECT id FROM dbsbuffer_file WHERE lfn = :lfn),
             (SELECT id FROM dbsbuffer_checksum_type WHERE type = :cktype), :cksum FROM dual"""

    def execute(self, lfn=None, cktype=None, cksum=None, bulkList=None, conn=None,
                transaction=False):

        if bulkList:
            binds = bulkList
        else:
            binds = [{'lfn': lfn, 'cktype': cktype, 'cksum': cksum}]

        self.dbi.processData(self.sql, binds,
                             conn=conn, transaction=transaction)

        return



"""
_ReleasePeriodicJob_

Oracle implementation of JobSplitting.ReleasePeriodicJob
"""

from WMCore.WMBS.MySQL.JobSplitting.ReleasePeriodicJob import ReleasePeriodicJob as MySQLReleasePeriodicJob

class ReleasePeriodicJob(MySQLReleasePeriodicJob):
    """
    Right now this is the same as the MySQL version.

    """



#!/usr/bin/env python
"""
_MarkOpen_

Oracle implementation of Fileset.MarkOpen
"""

__all__ = []



from WMCore.WMBS.MySQL.Fileset.MarkOpen import MarkOpen as MarkOpenFilesetMySQL

class MarkOpen(MarkOpenFilesetMySQL):
    sql = MarkOpenFilesetMySQL.sql



#!/usr/bin/env python
"""
_AddWorkUnits_

MySQL implementation of Jobs.AddWorkUnits
"""
from __future__ import absolute_import, division, print_function

import logging

from WMCore.Database.DBFormatter import DBFormatter


class AddWorkUnits(DBFormatter):
    """
    Add WorkUnit associations to jobs
    """

    sql = ('INSERT INTO wmbs_job_workunit_assoc (workunit, job) '
           ' SELECT wu.id, :jobid'
           ' FROM wmbs_workunit wu'
           ' INNER JOIN wmbs_frl_workunit_assoc frla ON wu.id = frla.workunit'
           ' WHERE frla.fileid=:fileid AND frla.run=:run AND frla.lumi=:lumi')

    def getBinds(self, jobFileRunLumis):
        """
        Translate a bulk list into a number of inserts

        Args:
            jobFileRunLumis: a list of tuples of the form [(jobid, fileid, run, lumi), ...]

        Returns: N/A
        """

        binds = []

        for jid, fid, run, lumi in jobFileRunLumis:
            binds.append({'jobid': jid, 'fileid': fid, 'run': run, 'lumi': lumi})

        return binds

    def execute(self, jobid=None, fileid=None, run=None, lumi=None, jobFileRunLumis=None, conn=None, transaction=False):
        """
        Args:
            jobid: The id of a single job
            fileid: The id of a single file
            run: Run # for a single work unit
            lumi: Lumi # for a single work unit
            jobFileRunLumis: a list of tuples of the form [(jobid, fileid, run, lumi), ...]

        Returns: N/A
        """

        if jobFileRunLumis:
            binds = self.getBinds(jobFileRunLumis)
        elif jobid and fileid and run and lumi:
            binds = DBFormatter.getBinds(self, jobid=jobid, fileid=fileid, run=run, lumi=lumi)
        elif not jobFileRunLumis and isinstance(jobFileRunLumis, list):  # Empty list is ok
            return
        else:
            logging.error('Jobs.AddWorkUnits called with insufficient arguments')
            return

        self.dbi.processData(self.sql, binds, conn=conn, transaction=transaction)
        return



"""
MySQL implementation of AddFile
"""
from WMCore.Database.DBFormatter import DBFormatter


class Add(DBFormatter):
    sql = """INSERT INTO wmbs_file_details (lfn, filesize, events,
                                            first_event, merged)
             VALUES (:lfn, :filesize, :events, :first_event,
                     :merged)"""

    def getBinds(self, files=None, size=0, events=0, cksum=0,
                 first_event=0, merged=False):
        # Can't use self.dbi.buildbinds here...
        binds = {}
        if not isinstance(files, (list, set)):
            binds = {'lfn': files,
                     'filesize': size,
                     'events': events,
                     'first_event': first_event,
                     'merged': int(merged)}
        else:
            # files is a list of tuples containing lfn, size, events, run and lumi
            binds = []
            for f in files:
                binds.append({'lfn': f[0],
                              'filesize': f[1],
                              'events': f[2],
                              'first_event': f[4],
                              'merged': int(f[5])
                              })
        return binds

    def execute(self, files=None, size=0, events=0, cksum=0,
                first_event=0, merged=False, conn=None,
                transaction=False):
        binds = self.getBinds(files, size, events, cksum, first_event,
                              merged)
        result = self.dbi.processData(self.sql, binds,
                                      conn=conn, transaction=transaction)
        return



#!/usr/bin/env python
"""
_UpdateDataset_

MySQL implementation of DBS3Buffer.UpdateDataset

Created on May 15, 2013

@author: dballest
"""

from WMCore.Database.DBFormatter import DBFormatter

class UpdateDataset(DBFormatter):
    """
    _UpdateDataset_

    Update the information of a dataset record in the
    DBSBuffer database
    """

    sql = """UPDATE dbsbuffer_dataset
              SET processing_ver = :procVer, acquisition_era = :acqEra,
                  valid_status = :valid, global_tag = :globalTag,
                  parent = :parent, prep_id = :prep_id
              WHERE id = :id
           """
    def execute(self, datasetId,
                processingVer = 0, acquisitionEra = None,
                validStatus = None, globalTag = None,
                parent = None, prep_id = None,
                conn = None, transaction = False):

        # to be backward compatible with various places in WMCore codebase which
        # assign processingVer to None default value. In DBS and WMCore databases
        # the processsing version should be integer data type
        if processingVer is None:
            processingVer = 0

        bindVars = {"procVer" : processingVer,
                    "acqEra" : acquisitionEra,
                    "valid" : validStatus,
                    "globalTag" : globalTag,
                    "parent" : parent,
                    "prep_id" : prep_id,
                    "id" : datasetId}
        self.dbi.processData(self.sql, bindVars, conn = conn,
                             transaction = transaction)

        return



from WMCore.WebTools.RESTModel import RESTModel
from cherrypy import response, request, HTTPError

class NestedModel(RESTModel):
    """
    A RESTModel that can also be used as a container for other RESTModels
    """

    def handler(self, verb, args=[], kwargs={}):
        """
        Call the appropriate method from self.methods for a given VERB. kwargs
        are query string parameters (e.g. method?thing1=abc&thing2=def).

        args are URI path elements, the first (arg[0]) is the base method name,
        other elements (arg[1:]) are either positional arguments to the base
        method or further components of the URI path, pointing at other methods.

        The default method should have a small number of arguments (preferably
        none). Any arguments passed to method beyond that number are checked to
        be path components, e.g. other method names.
        """
        verb = verb.upper()
        self._classifyHTTPError(verb, args, kwargs)
        args = list(args)
        basemethnom = args[0]
        basemethod = self.methods[verb][basemethnom]
        children = list(self.methods[verb][basemethnom])
        method = children.pop(children.index('default'))
        try:
            # is there a method in the keywords?
            for a in kwargs:
                if a in children:
                    method = a
                    if not len(kwargs[a]):
                        kwargs.pop(a)
            # is there a method in the positional args?
            for a in args[1:]:
                if a in children:
                    method = args.pop(args.index(a))

            data = basemethod[method]['call'](*args[1:], **kwargs)
        # in case sanitise_input is not called with in the method, if args doesn't
        # match throws the 400 error
        except TypeError as e:
            raise HTTPError(400, str(e))

        if 'expires' in basemethod[method]:
            return data, basemethod[method]['expires']
        else:
            return data, False



#!/usr/bin/env python

"""
_TestMonitor_

This is the test class for monitors
"""


import os.path
import time

from WMCore.Algorithms.SubprocessAlgos import *
from WMCore.WMRuntime.Monitors.WMRuntimeMonitor import WMRuntimeMonitor
from WMCore.WMSpec.Steps.Executor import getStepSpace
from WMCore.WMSpec.WMStep import WMStepHelper

getStepName = lambda step: WMStepHelper(step).name()


def getStepPID(stepSpace, stepName):
    """
    _getStepPID_

    Find the PID for a step given its stepSpace from the file
    """
    currDir = stepSpace.location
    pidFile = os.path.join(currDir, 'process_id')
    if not os.path.isfile(pidFile):
        msg = "Could not find process ID file for step %s" % stepName
        logging.error(msg)
        return

    with open(pidFile, 'r') as filehandle:
        output = filehandle.read()

    try:
        stepPID = int(output)
    except ValueError:
        msg = "Couldn't find a number"
        logging.error(msg)
        return None

    return stepPID


def searchForEvent(file):
    """
    _searchForEvent_

    Searches for the last event output into the CMSSW output file
    """

    MatchRunEvent = re.compile("Run: [0-9]+ Event: [0-9]+$")

    # I'm just grabbing the last twenty lines for the hell of it
    lines = tailNLinesFromFile(file, 20)

    lastMatch = None
    for line in lines:
        if MatchRunEvent.search(line.strip()):
            matches = MatchRunEvent.findall(line.strip())
            lastMatch = matches[-1]

    if lastMatch != None:
        #  //
        # // Extract and update last run/event number
        # //
        try:
            runInfo, lastEvent = lastMatch.split("Event:", 1)
            lastRun = int(runInfo.split("Run:", 1)[1])
            lastEvent = int(lastEvent)
            return (lastRun, lastEvent)
        except Exception:
            return (None, None)

    return (None, None)


class TestMonitor(WMRuntimeMonitor):
    def __init__(self):
        self.startTime = None
        self.currentStep = None
        self.currentStepName = None
        self.currentStepSpace = None
        self.softTimeOut = None
        self.hardTimeOut = None
        self.killFlag = False
        self.cmsswFile = None
        WMRuntimeMonitor.__init__(self)

    def initMonitor(self, task, job, logPath, args={}):
        """
        Handles the monitor initiation

        """
        print("In TestMonitor.initMonitor")

        self.softTimeOut = args.get('softTimeOut', None)
        self.hardTimeOut = args.get('hardTimeOut', None)

    def jobStart(self, task):
        """
        Job start notifier.
        """
        print("Yeehaw!  I started a job")

        return

    def jobEnd(self, task):
        """
        Job End notification

        """

        print("Job ended")

        return

    def stepStart(self, step):
        """
        Step start notification

        """
        self.currentStep = step
        self.currentStepName = getStepName(step)
        self.currentStepSpace = getStepSpace(self.currentStepName)
        self.startTime = time.time()
        print("Step started")

        return

    def stepEnd(self, step, stepReport):
        """
        Step end notification

        """
        self.currentStep = None
        self.currentStepName = None
        self.currentStepSpace = None
        print("Step ended")

    def periodicUpdate(self):
        """
        Run on the defined intervals.

        """

        if not self.currentStep or not self.currentStepSpace:
            # We're probably between steps
            return

        # Check for events
        if self.cmsswFile:
            run, event = searchForEvent(self.cmsswFile)
            if run and event:
                # Then we actually found something, otherwise do nothing
                # Right now I don't know what to do
                pass

        # Do timeout
        if not self.softTimeOut:
            return

        if time.time() - self.startTime > self.softTimeOut:
            # Then we have to kill the process

            # First, get the PID
            stepPID = getStepPID(self.currentStepSpace, self.currentStepName)

            # Now kill it!
            msg = ""
            msg += "Start Time: %s\n" % self.startTime
            msg += "Time Now: %s\n" % time.time()
            msg += "Timeout: %s\n" % self.softTimeOut
            msg += "Killing Job...\n"
            msg += "Process ID is: %s\n" % stepPID
            if time.time() - self.startTime < self.hardTimeOut or not self.killFlag:
                msg += "WARNING: Soft Kill Timeout has Expired:"
                logging.error(msg)
                os.kill(stepPID, signal.SIGUSR2)
                self.killFlag = True
            elif self.killFlag:
                msg += "WARNING: Hard Kill Timeout has Expired:"
                logging.error(msg)
                os.kill(stepPID, signal.SIGTERM)
                killedpid, stat = os.waitpid(stepPID, os.WNOHANG)
                if killedpid == 0:
                    os.kill(stepPID, signal.SIGKill)
                    killedpid, stat = os.waitpid(stepPID, os.WNOHANG)
                    if killedpid == 0:
                        logging.error("Can't kill job.  Out of options.  Waiting for system reboot.")
                        # Panic!  It's unkillable!
                        pass


                        # logging.error(msg)



#!/usr/bin/env python
"""
_GetParentStatus_

MySQL implementation of DBSBufferFile.GetParentStatus
"""




from WMCore.Database.DBFormatter import DBFormatter

class GetParentStatus(DBFormatter):
    sql = """SELECT status FROM dbsbuffer_file
               INNER JOIN dbsbuffer_file_parent ON
                 dbsbuffer_file.id = dbsbuffer_file_parent.parent
             WHERE dbsbuffer_file_parent.child =
               (SELECT id FROM dbsbuffer_file WHERE lfn  = :lfn)"""

    def format(self, results):
        """
        _format_

        Format the query results into a list of LFNs.
        """
        results = DBFormatter.format(self, results)

        status = []
        for result in results:
            status.append(result[0])
        return status

    def execute(self, lfn, conn = None, transaction = False):
        result = self.dbi.processData(self.sql, {"lfn": lfn}, conn = conn,
                                      transaction = transaction)
        return self.format(result)



"""
__UpdatePriority__

MySQL implementation of UpdatePriority

Created on Apr 16, 2013

@author: dballest
"""

from WMCore.Database.DBFormatter import DBFormatter

class UpdatePriority(DBFormatter):
    """
    _UpdatePriority_

    Update the priority of the workflows given
    the name.
    """
    sql = """UPDATE wmbs_workflow SET priority = :priority
             WHERE wmbs_workflow.name = :name"""

    def execute(self, workflowsToUpdate,
                conn = None, transaction = False):
        binds = []
        for workflow in workflowsToUpdate:
            bind = {'name' : workflow,
                    'priority' : workflowsToUpdate[workflow]}
            binds.append(bind)
        self.dbi.processData(self.sql, binds,
                             conn = conn, transaction = transaction)



#!/usr/bin/env python
"""
MySQL implementation of AddFile
"""

from builtins import str, bytes

from WMCore.Database.DBFormatter import DBFormatter


class Add(DBFormatter):
    sql = """INSERT IGNORE dbsbuffer_file(lfn, filesize, events, dataset_algo, status, workflow, in_phedex)
                values (:lfn, :filesize, :events, :dataset_algo, :status, :workflow, :in_phedex)"""

    def getBinds(self, files, size, events, cksum, dataset_algo, status, workflowID, inPhedex):
        # Can't use self.dbi.buildbinds here...
        binds = []
        if isinstance(files, (str, bytes)):
            bind = {'lfn': files,
                    'filesize': size,
                    'events': events,
                    'dataset_algo': dataset_algo,
                    'status': status,
                    'workflow': workflowID,
                    'in_phedex': inPhedex}
            binds.append(bind)
        elif isinstance(files, (list, set)):
            # files is a list of tuples containing lfn, size, events, cksum, dataset, status
            for f in files:
                binds.append({'lfn': f[0],
                              'filesize': f[1],
                              'events': f[2],
                              'dataset_algo': f[3],
                              'status': f[4],
                              'workflow': f[5],
                              'in_phedex': f[6]})
        return binds

    def execute(self, files=None, size=0, events=0, cksum=0,
                datasetAlgo=0, status="NOTUPLOADED", workflowID=None,
                inPhedex=0, conn=None, transaction=False):
        binds = self.getBinds(files, size, events, cksum, datasetAlgo, status,
                              workflowID, inPhedex)

        self.dbi.processData(self.sql, binds, conn=conn, transaction=transaction)

        return



#! /usr/bin/env python

from builtins import str, map
import collections
from itertools import islice, chain, groupby

def grouper(iterable, n):
    """
    :param iterable: List of other iterable to slice
    :type: iterable
    :param n: Chunk size for resulting lists
    :type: int
    :return: iterator of the sliced list

    Source: http://stackoverflow.com/questions/3992735/python-generator-that-groups-another-iterable-into-groups-of-n
    """
    iterable = iter(iterable)
    return iter(lambda: list(islice(iterable, n)), [])


def getChunk(arr, step):
    """
    Return chunk of entries from given array and step, it is similar in behavior to grouper
    function but instead of returning new list it provides a generator iterable object.
    :param arr: input array of data
    :param step: step to iterate
    :return: generator, set of slices with number of entries equal to step of iteration
    """
    for i in range(0, len(arr), step):
        yield arr[i:i + step]


def flattenList(doubleList):
    """
    Make flat a list of lists.
    """
    return list(chain.from_iterable(doubleList))


def nestedDictUpdate(d, u):
    """
    Code from Alex Matelli
    http://stackoverflow.com/questions/3232943/update-value-of-a-nested-dictionary-of-varying-depth
    """
    for k, v in u.items():
        if isinstance(v, collections.Mapping):
            r = nestedDictUpdate(d.get(k, {}), v)
            d[k] = r
        else:
            d[k] = u[k]
    return d

def convertFromUnicodeToBytes(data):
    """
    code fram
    http://stackoverflow.com/questions/1254454/fastest-way-to-convert-a-dicts-keys-values-from-unicode-to-str
    """
    if isinstance(data, str):
        return data.encode('utf-8')
    elif isinstance(data, collections.Mapping):
        return dict(list(map(convertFromUnicodeToBytes, list(data.items()))))
    elif isinstance(data, collections.Iterable):
        return type(data)(list(map(convertFromUnicodeToBytes, data)))
    else:
        return data


def makeListElementsUnique(listObj):
    """
    Given a list of lists or a list of tuples, find all duplicate elements
    and make them unique.
    :param listObj: an unsorted list of lists or a list of tuples, e.g.:
        [[1, 1], [1, 5], [1, 1]]; or
        [(1, 1), (1, 5), (1, 1)]
    :return: the same list object but with no duplicates

    Source: https://stackoverflow.com/questions/2213923/removing-duplicates-from-a-list-of-lists
    """
    listObj.sort()
    return list(k for k, _ in groupby(listObj))



"""
Use this for only unit test
"""
from builtins import object

class EmulatorHelper(object):
    """
    Works as global value for the emulator switch.
    WARNNING: This is not multi thread safe.
    """
    #DO not change default values
    LocalCouchDBData = None
    ReqMonDBData = None
    WMAgentDBData= None

    @staticmethod
    def getEmulatorClass(clsName):

        if clsName == 'LocalCouchDBData':
            from WMQuality.Emulators.AnalyticsDataCollector.DataCollectorAPI \
                import LocalCouchDBData as LocalCouchDBDataEmulator
            return LocalCouchDBDataEmulator

        if clsName == 'ReqMonDBData':
            from WMQuality.Emulators.AnalyticsDataCollector.DataCollectorAPI \
                import ReqMonDBData as ReqMonDBDataEmulator
            return ReqMonDBDataEmulator

        if clsName == 'WMAgentDBData':
            from WMQuality.Emulators.AnalyticsDataCollector.DataCollectorAPI \
                import WMAgentDBData as WMAgentDBDataEmulator
            return WMAgentDBDataEmulator

    @staticmethod
    def setEmulators(localCouch=False,reqMon=False, wmagentDB=False):
        EmulatorHelper.LocalCouchDBData = localCouch
        EmulatorHelper.ReqMonDBData = reqMon
        EmulatorHelper.WMAgentDBData = wmagentDB

    @staticmethod
    def resetEmulators():
        EmulatorHelper.LocalCouchDBData = None
        EmulatorHelper.ReqMonDBData = None
        EmulatorHelper.WMAgentDBData = None

    @staticmethod
    def getClass(cls):
        """
        if emulator flag is set return emulator class
        otherwise return original class.
        if emulator flag is not initialized
            and EMULATOR_CONFIG environment variable is set,
        """
        emFlag = getattr(EmulatorHelper, cls.__name__)
        if emFlag:
            return EmulatorHelper.getEmulatorClass(cls.__name__)
        elif emFlag == None:
            try:
                from WMQuality.Emulators import emulatorSwitch
            except:
                # if emulatorSwitch class is not imported don't use
                # emulator
                setattr(EmulatorHelper, cls.__name__, False)
            else:
                envFlag = emulatorSwitch(cls.__name__)
                setattr(EmulatorHelper, cls.__name__, envFlag)
                if envFlag:
                    return EmulatorHelper.getEmulatorClass(cls.__name__)
        # if emulator flag is False, return original class
        return cls

def emulatorHook(cls):
    """
    This is used as decorator to switch between Emulator and real Class
    on instance creation.
    """
    class EmulatorWrapper(object):
        def __init__(self, *args, **kwargs):
            aClass = EmulatorHelper.getClass(cls)
            self.wrapped = aClass(*args, **kwargs)
            self.__class__.__name__ = self.wrapped.__class__.__name__

        def __getattr__(self, name):
            return getattr(self.wrapped, name)

    return EmulatorWrapper



#!/usr/bin/env python
"""
_KillWorkflow_

Oracle implementation of Subscriptions.KillWorkflow
"""

from WMCore.WMBS.MySQL.Subscriptions.KillWorkflow import KillWorkflow as MySQLKillWorkflow

class KillWorkflow(MySQLKillWorkflow):
    """
    _KillWorkflow_

    Mark all files that are not complete/failed and belong to a particular
    workflow as failed.  Ignore Cleanup and LogCollect subscriptions as we
    still want those to run.
    """
    pass



"""
_ExistWorker_

MySQL implementation of ExistWorker
"""

__all__ = []



import time
from WMCore.Database.DBFormatter import DBFormatter

class ExistWorker(DBFormatter):

    sql = """SELECT component_id FROM wm_workers
                WHERE component_id = (SELECT id FROM wm_components
                                   WHERE name = :component_name)
                AND name = :worker_name
             """

    def format(self, result):
        result = DBFormatter.format(self, result)

        if len(result) == 0:
            return False
        else:
            return result[0][0]

    def execute(self, componentName, workerName,
                conn = None, transaction = False):

        binds = {"component_name": componentName,
                 "worker_name": workerName}

        result = self.dbi.processData(self.sql, binds, conn = conn,
                             transaction = transaction)
        return self.format(result)



#!/usr/bin/env python
"""
_InsertType_

Oracle implementation of Subscription.InsertType
"""




from WMCore.WMBS.MySQL.Subscriptions.InsertType import InsertType as InsertTypeMySQL

class InsertType(InsertTypeMySQL):
    sql = """INSERT INTO wmbs_sub_types (id, name)
               SELECT wmbs_sub_types_SEQ.nextval, :name AS name FROM DUAL WHERE NOT EXISTS
                (SELECT name FROM wmbs_sub_types WHERE name = :name)"""



#!/usr/bin/env python
"""
_Harvest_

"""
from future.utils import viewitems

import threading
import logging

from WMCore.JobSplitting.JobFactory import JobFactory
from WMCore.Services.UUIDLib import makeUUID
from WMCore.DAOFactory import DAOFactory
from WMCore.JobSplitting.LumiBased import isGoodRun, isGoodLumi
from WMCore.DataStructs.Run import Run
from WMCore.WMSpec.WMTask import buildLumiMask


class Harvest(JobFactory):
    """
    _Harvest_

    Job splitting algoritm which creates a single job for all files
    in the fileset (not neccessarily just available files).
    Two distinct modes, Periodic and EndOfRun.

    In Periodic mode, we periodically create a job processing all
    files. A job will not be created until the previous job (if
    there is one) has been completed and there are new available
    files in the fileset. The specified period is the amount of
    time in seconds between the end of a job and the creation of
    another job.

    In EndOfRun mode, create a job processing all files once the
    input file has been closed. This means there will only be
    a single job in total for the subscription.

    For the EndOfRun mode support a sibling parameters that is
    set if there is also a Periodic subscription. In this case
    wait until the Periodic subscription is finished before
    triggering the EndOfRun harvesting.

    """

    def createJobsLocationWise(self, fileset, endOfRun, dqmHarvestUnit, lumiMask, goodRunList):

        myThread = threading.currentThread()
        fileset.loadData(parentage=0)
        allFiles = fileset.getFiles()

        # sort by location and run
        locationDict = {}
        runDict = {}
        for fileInfo in allFiles:

            locSet = frozenset(fileInfo['locations'])
            runSet = fileInfo.getRuns()

            if len(locSet) == 0:
                logging.error("File %s has no locations!", fileInfo['lfn'])
            if len(runSet) == 0:
                logging.error("File %s has no run information!", fileInfo['lfn'])

            # Populate a dictionary with [location][run] so we can split jobs according to those different combinations
            if locSet not in locationDict:
                locationDict[locSet] = {}

            fileInfo['runs'] = set()
            # Handle jobs with run whitelist/blacklist
            if goodRunList:
                runDict[fileInfo['lfn']] = set()
                for run in runSet:
                    if run.run in goodRunList:
                        runDict[fileInfo['lfn']].add(run)
                        if run.run in locationDict[locSet]:
                            locationDict[locSet][run.run].append(fileInfo)
                        else:
                            locationDict[locSet][run.run] = [fileInfo]
            elif lumiMask:
                # it has lumiMask, thus we consider only good run/lumis
                newRunSet = []
                for run in runSet:
                    if not isGoodRun(lumiMask, run.run):
                        continue
                    # then loop over lumis
                    maskedLumis = []
                    for lumi in run.lumis:
                        if not isGoodLumi(lumiMask, run.run, lumi):
                            continue
                        maskedLumis.append(lumi)

                    if not maskedLumis:
                        continue
                    maskedRun = Run(run.run, *maskedLumis)
                    newRunSet.append(maskedRun)

                    if run.run in locationDict[locSet]:
                        locationDict[locSet][run.run].append(fileInfo)
                    else:
                        locationDict[locSet][run.run] = [fileInfo]
                if newRunSet:
                    runDict[fileInfo['lfn']] = newRunSet
            else:
                # no LumiList and no run white or black list
                runDict[fileInfo['lfn']] = runSet
                for run in runSet:
                    if run.run in locationDict[locSet]:
                        locationDict[locSet][run.run].append(fileInfo)
                    else:
                        locationDict[locSet][run.run] = [fileInfo]

        # create separate jobs for different locations
        self.newGroup()
        self.jobCount = 0
        baseName = makeUUID()
        self.newGroup()

        if endOfRun:
            harvestType = "EndOfRun"
        else:
            harvestType = "Periodic"

        for location in locationDict:

            if dqmHarvestUnit == "byRun":
                self.createJobByRun(locationDict, location, baseName, harvestType, runDict, endOfRun)
            else:
                self.createMultiRunJob(locationDict, location, baseName, harvestType, runDict, endOfRun)

        return

    def createJobByRun(self, locationDict, location, baseName, harvestType, runDict, endOfRun):
        """
        _createJobByRun_

        Creates one job per run for all files available at the same location.
        """

        for run in locationDict[location]:
            # Should create at least one job for every location/run, putting this here will do
            self.jobCount += 1
            self.newJob(name="%s-%s-Harvest-%i" % (baseName, harvestType, self.jobCount))
            for f in locationDict[location][run]:
                for fileRun in runDict[f['lfn']]:
                    if fileRun.run == run:
                        self.currentJob['mask'].addRun(fileRun)
                        break
                self.currentJob.addFile(f)

            if endOfRun:
                self.currentJob.addBaggageParameter("runIsComplete", True)
            self.mergeLumiRange(self.currentJob['mask']['runAndLumis'])
        return

    def createMultiRunJob(self, locationDict, location, baseName, harvestType, runDict, endOfRun):
        """
        _createMultiRunJob_

        Creates a single harvesting job for all files and runs available
        at the same location.
        """

        self.jobCount += 1
        self.newJob(name="%s-%s-Harvest-%i" % (baseName, harvestType, self.jobCount))
        for run in locationDict[location]:
            for f in locationDict[location][run]:
                for fileRun in runDict[f['lfn']]:
                    if fileRun.run == run:
                        self.currentJob['mask'].addRun(fileRun)
                        break
                if f not in self.currentJob['input_files']:
                    self.currentJob.addFile(f)

        if endOfRun:
            self.currentJob.addBaggageParameter("runIsComplete", True)
        self.mergeLumiRange(self.currentJob['mask']['runAndLumis'])

        # now calculate the minimum and maximum run number, it has to go to the root name
        minRun = min(self.currentJob['mask']['runAndLumis'].keys())
        maxRun = max(self.currentJob['mask']['runAndLumis'].keys())

        self.currentJob.addBaggageParameter("multiRun", True)
        self.currentJob.addBaggageParameter("runLimits", "-%s-%s" % (minRun, maxRun))

        return

    def mergeLumiRange(self, runLumis):
        """
        _mergeLumiRange_

        Merges the interesection of lumi ranges.
        """
        for run, lumis in viewitems(runLumis):
            lumis.sort(key=lambda sublist: sublist[0])
            fixedLumis = [lumis[0]]
            for lumi in lumis:
                if (fixedLumis[-1][1] + 1) >= lumi[0]:
                    fixedLumis[-1][1] = lumi[1]
                else:
                    fixedLumis.append(lumi)
            self.currentJob['mask']['runAndLumis'][run] = fixedLumis

    def algorithm(self, *args, **kwargs):
        """
        _algorithm_

        """

        myThread = threading.currentThread()

        periodicInterval = kwargs.get("periodic_harvest_interval", 0)
        periodicSibling = kwargs.get("periodic_harvest_sibling", False)
        dqmHarvestUnit = kwargs.get("dqmHarvestUnit", "byRun")
        runs = kwargs.get("runs", None)
        lumis = kwargs.get("lumis", None)
        runWhitelist = set(kwargs.get('runWhitelist', []))
        runBlacklist = set(kwargs.get('runBlacklist', []))
        goodRunList = runWhitelist.difference(runBlacklist)

        daoFactory = DAOFactory(package="WMCore.WMBS",
                                logger=myThread.logger,
                                dbinterface=myThread.dbi)

        releasePeriodicJobDAO = daoFactory(classname="JobSplitting.ReleasePeriodicJob")
        periodicSiblingCompleteDAO = daoFactory(classname="JobSplitting.PeriodicSiblingComplete")

        fileset = self.subscription.getFileset()
        fileset.load()

        lumiMask = {}
        if runs and lumis:
            lumiMask = buildLumiMask(runs, lumis)

        if periodicInterval and periodicInterval > 0:

            # Trigger the Periodic Job if
            #  * it is the first job OR
            #  * the last job ended more than periodicInterval seconds ago
            triggerJob = releasePeriodicJobDAO.execute(subscription=self.subscription["id"], period=periodicInterval)

            if triggerJob:
                myThread.logger.debug("Creating Periodic harvesting job")
                self.createJobsLocationWise(fileset, False, dqmHarvestUnit, lumiMask, goodRunList)

        elif not fileset.open:

            # Trigger the EndOfRun job if
            #  * (same as Periodic to not have JobCreator go nuts and stop after the first iteration)
            #  * there is no Periodic sibling subscription OR
            #  * the Periodic sibling subscription is complete
            triggerJob = releasePeriodicJobDAO.execute(subscription=self.subscription["id"], period=3600)
            if triggerJob and periodicSibling:
                triggerJob = periodicSiblingCompleteDAO.execute(subscription=self.subscription["id"])

            if triggerJob:
                myThread.logger.debug("Creating EndOfRun harvesting job")
                self.createJobsLocationWise(fileset, True, dqmHarvestUnit, lumiMask, goodRunList)

        return



#!/usr/bin/env python
"""
_Scram_

Command line scram environment wrapper that can be used to run small commands
requiring the scram environment setup

sample usage:

    scram = Scram(
        version = "CMSSW_X_Y_Z",             # CMSSW Version
        directory = "path to working area"   # defaults to cwd
        initialise = " . /uscmst1/prod/sw/cms/setup/shrc prod"  # boot strap env command
    )



    scram.project() # creates project area
    scram.runtime() # creates runtime environment

    # run something requireing scram env
    retval = scram("edmFileUtil")
    print retval, scram.stdout, scram.stderr


"""

from builtins import range, object

import logging
import os
import os.path
import subprocess
import sys
import platform

from PSetTweaks.WMTweak import readAdValues
from Utils.Utilities import encodeUnicodeToBytes, decodeBytesToUnicode

SCRAM_TO_ARCH = {'amd64': 'X86_64', 'aarch64': 'aarch64', 'ppc64le': 'ppc64le'}
# Scram arch to platform machine values above are unique, so we can reverse the mapping
ARCH_TO_SCRAM = {arch:scram for scram, arch in list(SCRAM_TO_ARCH.items())}
ARCH_TO_OS = {'slc5': ['rhel6'],
              'slc6': ['rhel6'],
              'slc7': ['rhel7'],
              'el8': ['rhel8'], 'cc8': ['rhel8'], 'cs8': ['rhel8'], 'alma8': ['rhel8'],
              'el9': ['rhel9'], 'cs9': ['rhel9'],
              }
OS_TO_ARCH = {}
for arch, oses in ARCH_TO_OS.items():
    for osName in oses:
        if osName not in OS_TO_ARCH:
            OS_TO_ARCH[osName] = []
        OS_TO_ARCH[osName].append(arch)

def getPlatformMachine():
    """
    Get the platform machine. Keep it compatible with HTCondor
    :return: str, platform machine.
    """
    machine = platform.machine()
    # Condor throws X86_64 but machine uses x86_64. Using condor convention
    if machine == 'x86_64':
        machine = machine.upper()

    return machine

def getSingleScramArch(scramArch):
    """
    Figure out which scram arch is compatible with both the request and the release

    Args:
        scramArch: string or list of strings representing valid scram arches for the workflow

    Returns:
        a single scram arch
    """
    if isinstance(scramArch, list):
        try:
            ad = readAdValues(['glidein_required_os'], 'machine')
            runningOS = ad['glidein_required_os'].strip('"')
            validArches = sorted(OS_TO_ARCH[runningOS], reverse=True)
            for requestedArch in sorted(scramArch, reverse=True):
                for validArch in validArches:
                    if requestedArch.startswith(validArch):
                        # Check target machine matches with a valid scram arch
                        if ARCH_TO_SCRAM.get(getPlatformMachine()):
                            return requestedArch
        except KeyError:
            return sorted(scramArch)[-1]  # Give the most recent release if lookup fails
        return None
    else:
        return scramArch


def isCMSSWSupported(thisCMSSW, supportedCMSSW):
    """
    _isCMSSWSupported_

    Function used to validate whether the CMSSW release to be used supports
    a feature that is not available in all releases.
    :param thisCMSSW: release to be used in this job
    :param allowedCMSSW: first (lowest) release that started supporting the
    feature you'd like to use.
    
    NOTE: only the 3 digits version are evaluated, pre and patch releases
    are not taken into consideration
    """
    if not thisCMSSW or not supportedCMSSW:
        logging.info("You must provide the CMSSW version being used by this job and a supported version")
        return False

    if thisCMSSW == supportedCMSSW:
        return True

    thisCMSSW = [int(i) for i in thisCMSSW.split('_', 4)[1:4]]
    supportedCMSSW = [int(i) for i in supportedCMSSW.split('_', 4)[1:4]]
    for idx in range(3):
        if thisCMSSW[idx] > supportedCMSSW[idx]:
            return True
        elif thisCMSSW[idx] == supportedCMSSW[idx] and idx < 2:
            if thisCMSSW[idx + 1] > supportedCMSSW[idx + 1]:
                return True
        else:
            return False

    return False


def isEnforceGUIDInFileNameSupported(thisCMSSW):
    """
    _isEnforceGUIDInFileNameSupported_

    Function used to validate whether the CMSSW release to be used supports
    the enforceGUIDInFileName feature.
    :param thisCMSSW: release to be used in this job

    """
    # a set of CMSSW releases that support the enforceGUIDInFileName feature. Releases in the same
    # cycle with a higher minor revision number also support the feature.
    supportedReleases = set(["CMSSW_10_6_8", "CMSSW_10_2_20", "CMSSW_9_4_16", "CMSSW_9_3_17", "CMSSW_8_0_34"])
    # a set of specific CMSSW releases that supported the enforceGUIDInFileName feature.
    specificSupportedReleases = set(["CMSSW_10_2_20_UL", "CMSSW_9_4_16_UL", "CMSSW_8_0_34_UL", "CMSSW_7_1_45_patch3"])

    if not thisCMSSW:
        logging.info("You must provide the CMSSW version being used by this job.")
        return False

    # true if CMSSW release is >= CMSSW_11_0_0
    if isCMSSWSupported(thisCMSSW, "CMSSW_11_0_0"):
        return True
    # true if CMSSW release is in the specific release set
    elif thisCMSSW in specificSupportedReleases:
        return True
    # true if the CMSSW release's minor revision is >= to one of the supported releases
    else:
        thisMajor, thisMid, thisMinor = [int(i) for i in thisCMSSW.split('_', 4)[1:4]]
        for release in supportedReleases:
            supportedMajor, supportedMid, supportedMinor = [int(i) for i in release.split('_', 4)[1:4]]
            # major and mid revisions need an exact match
            if thisMajor == supportedMajor and thisMid == supportedMid:
                # minor revision need to be >= to the supported minor revision
                if thisMinor >= supportedMinor:
                    return True
    return False


def testWriter(func, *args):
    """
    _testWriter_

    If running Scram object in unittests, this mode will neuter the shell commands to make them
    echo the commands written to the subprocess.

    """

    def newWriter(*args):
        newWriter.__name__ = func.__name__
        newWriter.__doc__ = func.__doc__
        newWriter.__dict__.update(func.__dict__)
        subproc = args[0]
        line = args[1]
        escapedLine = "echo \"%s\"\n" % line
        func(subproc, escapedLine)

    return newWriter


#  //
# // Interceptable function to push commands to the subshell, used to
# //  enable test mode.
procWriter = lambda s, l: s.stdin.write(encodeUnicodeToBytes(l))


class Scram(object):
    """
    _Scram_

    Object to encapsulate a scram "session" that can be used to create
    a project area, bootstrap the environment and then use that to
    execute commands

    Simple enumeration of scram errors is performed, to allow to be mapped to standard
    exception/error conditions

    """

    def __init__(self, **options):
        self.command = options.get("command", "scramv1")
        self.initialise = options.get("initialise", None)
        self.version = options.get("version", None)
        self.directory = options.get("directory", os.getcwd())
        self.architecture = options.get("architecture", None)
        self.test_mode = options.get("test", False)
        self.envCmd = options.get("envCmd", None)

        # state checks
        self.projectArea = None
        self.runtimeEnv = {}

        # handler to write to subprocesses
        self.library_path = "export LD_LIBRARY_PATH=$LD_LIBRARY_PATH\n"
        self.procWriter = procWriter
        if self.test_mode:
            # if in test mode, decorate the subprocess writer with the test harness
            self.procWriter = testWriter(procWriter)
            # dont actually try to call a non-existent scram binary in test mode
            self.command = "/bin/echo"

        # buffers for debug/error reporting
        self.stdout = None
        self.stderr = None
        self.code = None
        self.lastExecuted = None

    def preCommand(self):
        """
        _preCommand_

        build the pre execution command for scram environment

        """
        result = ""
        if self.architecture is not None:
            result += "export SCRAM_ARCH=%s\n" % self.architecture

        if self.initialise is not None:
            result += "%s\n" % self.initialise
            result += """if [ "$?" -ne "0" ]; then exit 2; fi\n"""
        return result

    def project(self):
        """
        _project_

        Do a scram project command in the directory provided
        Sets the projectArea attribute checked by runtime and call

        """
        proc = subprocess.Popen(
            ["/bin/bash"], shell=True, cwd=self.directory,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            stdin=subprocess.PIPE,
        )

        # send commands to the subshell utilising the process writer method
        self.procWriter(proc, self.preCommand())
        self.procWriter(proc, "%s --arch %s project CMSSW %s\n" % (self.command, self.architecture, self.version))
        self.procWriter(proc, """if [ "$?" -ne "0" ]; then exit 3; fi\n""")
        self.procWriter(proc, "exit 0")

        self.projectArea = "%s/%s" % (self.directory, self.version)
        self.stdout, self.stderr = proc.communicate()
        self.code = proc.returncode
        self.lastExecuted = "%s project CMSSW %s" % (
            self.command, self.version)
        if self.test_mode:
            # in test mode, the scram command would create the project area
            # have to emulate this by hand here.
            if not os.path.exists(self.projectArea):
                os.makedirs(self.projectArea)
        return proc.returncode

    def runtime(self):
        """
        _runtime_

        Scram runtime command

        """
        if self.projectArea is None:
            msg = "Scram Runtime called with Project Area not set"
            self.stdout = msg
            self.stderr = ""
            return 1

        try:
            proc = subprocess.Popen(
                ["/bin/bash"], shell=True, cwd=self.projectArea,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                stdin=subprocess.PIPE,
            )
        except Exception as ex:
            msg = "Error thrown while invoking subprocess for scram runtime\n"
            msg += "%s\n" % str(ex)
            msg += "Opening subprocess shell in %s" % self.projectArea
            self.stdout = msg
            self.stderr = ""
            return 1

        # write via process writer method
        self.procWriter(proc, self.library_path)
        self.procWriter(proc, self.preCommand())
        self.procWriter(proc, "%s ru -sh\n" % self.command)
        self.procWriter(proc, """if [ "$?" -ne "0" ]; then exit 4; fi\n""")
        self.procWriter(proc, "eval `%s ru -sh`\n" % self.command)

        self.stdout, self.stderr = proc.communicate()
        self.stdout = decodeBytesToUnicode(self.stdout)
        self.stderr = decodeBytesToUnicode(self.stderr)
        if proc.returncode == 0:
            for l in self.stdout.split(";\n"):
                if l.strip() == "":
                    continue
                if l.strip().startswith('unset'):
                    continue
                l = l.replace("export ", "")
                try:
                    var, val = l.split("=", 1)
                except ValueError as ex:
                    raise ValueError("Couldn't split line: %s" % l) from ex

                self.runtimeEnv[var] = val
        if self.test_mode:
            # ensure that runtime env isnt empty in test mode
            # as that will cause an exception in __call__
            self.runtimeEnv['TEST_MODE'] = 1
        self.code = proc.returncode
        self.lastExecuted = "eval `%s ru -sh`" % self.command
        return proc.returncode

    def __call__(self, command, hackLdLibPath=False, runtimeDir=None, cleanEnv=True):
        """
        _operator(command)_

        Run the command in the runtime environment

        """
        if self.projectArea is None:
            msg = "Scram Project Area not set/project() not called"
            raise RuntimeError(msg)
        if self.runtimeEnv == {}:
            msg = "Scram runtime environment is empty/runtime() not called"
            raise RuntimeError(msg)
        executeIn = runtimeDir
        if runtimeDir is None:
            executeIn = self.projectArea

        bashcmd = "/bin/bash"
        if cleanEnv:
            # Start with a clean environment
            bashcmd = "env - " + bashcmd

        logging.info("Creating a subprocess to run the PSet setup.")
        logging.info("Also recording SCRAM command-line related output.")
        proc = subprocess.Popen([bashcmd], shell=True, cwd=executeIn,
                                stdout=subprocess.PIPE,
                                stderr=subprocess.STDOUT,
                                stdin=subprocess.PIPE,
                                )

        # Passing the environment in to the subprocess call results in all of
        # the variables being quoted which causes problems for search paths.
        # We'll setup the environment the hard way to avoid this.
        rtCmsswBase = None
        rtScramArch = self.architecture
        for varName in self.runtimeEnv:
            self.procWriter(proc, 'export %s=%s\n' % (varName, self.runtimeEnv[varName]))
            if varName == "CMSSW_RELEASE_BASE":
                rtCmsswBase = self.runtimeEnv[varName].replace('\"', '')
            if varName == "SCRAM_ARCH":
                rtScramArch = self.runtimeEnv[varName].replace('\"', '')

        if os.environ.get('VO_CMS_SW_DIR', None) is not None:
            self.procWriter(proc, 'export VO_CMS_SW_DIR=%s\n' % os.environ['VO_CMS_SW_DIR'])
        if os.environ.get('OSG_APP', None) is not None:
            self.procWriter(proc, 'export VO_CMS_SW_DIR=%s/cmssoft/cms\n' % os.environ['OSG_APP'])
        # In general, CMSSW releases <= 12_6_0 will use CMS_PATH, while anything beyond that
        # requires the SITECONFIG_PATH variable to properly load storage.xml/json file.
        if os.environ.get('CMS_PATH', None) is not None:
            self.procWriter(proc, 'export CMS_PATH=%s\n' % os.environ['CMS_PATH'])
        if os.environ.get('SITECONFIG_PATH', None) is not None:
            self.procWriter(proc, 'export SITECONFIG_PATH=%s\n' % os.environ['SITECONFIG_PATH'])
        if os.environ.get('_CONDOR_JOB_AD'):
            self.procWriter(proc, 'export _CONDOR_JOB_AD=%s\n' % os.environ['_CONDOR_JOB_AD'])
        if os.environ.get('_CONDOR_MACHINE_AD'):
            self.procWriter(proc, 'export _CONDOR_MACHINE_AD=%s\n' % os.environ['_CONDOR_MACHINE_AD'])

        if hackLdLibPath:
            self.procWriter(proc, self.library_path)

        self.procWriter(proc, "%s\n" % self.preCommand())

        if self.envCmd is not None:
            self.procWriter(proc, "%s\n" % self.envCmd)

        if command.startswith(sys.executable):
            # replace COMP python version with python from CMSSW release if possible
            python27exec = "%s/external/%s/bin/python2.7" % (rtCmsswBase, rtScramArch)
            python26exec = "%s/external/%s/bin/python2.6" % (rtCmsswBase, rtScramArch)
            if os.path.islink(python27exec):
                command = command.replace(sys.executable, python27exec)
            elif os.path.islink(python26exec):
                command = command.replace(sys.executable, python26exec)
            elif hackLdLibPath:
                # reset python path for DMWM python (scram will have changed env to point at its own)
                self.procWriter(proc, "export PYTHONPATH=%s:$PYTHONPATH\n" % ":".join(sys.path)[1:])

        logging.info("    Invoking command: %s", command)
        self.procWriter(proc, "%s\n" % command)
        self.procWriter(proc, """if [ "$?" -ne "0" ]; then exit 5; fi\n""")
        self.stdout, self.stderr = proc.communicate()
        self.stdout = decodeBytesToUnicode(self.stdout)
        self.stderr = decodeBytesToUnicode(self.stderr)
        logging.info("Subprocess stdout was:\n%s", self.stdout)
        logging.info("Subprocess stderr was:\n%s", self.stderr)
        self.code = proc.returncode
        self.lastExecuted = command

        return self.code

    def diagnostic(self):
        """
        _diagnostic_

        Diagnostic Error message for a scram command failure

        """

        result = """
        Scram Command Diagnostic:
        Command : %s
        Architecture: %s
        Executed: %s
        Exit Status: %s
        Stdout: %s
        Stderr: %s """ % (
            self.command,
            self.architecture,
            self.lastExecuted,
            self.code,
            self.stdout,
            self.stderr)
        return result

    def getStdout(self):
        """
        retrieve stdout of the command executed via the __call__ method\
        note that in current implementation stderr of command execution
        is piped to stdout
        :return: a (possibly long) string
        """
        return self.stdout



#!/usr/bin/env python
from WMCore.WebTools.Page import exposexml, exposejson
from WMCore.WebTools.Page import exposedasplist, exposedasjson, exposedasxml
from WMCore.WebTools.DatabasePage import DatabasePage
from Utils.Utilities import encodeUnicodeToBytes
from cherrypy import expose, HTTPRedirect
from Utils.CPMetrics import promMetrics
import sys
import traceback

# third-party libraries
from cherrypy.lib import cpstats

class WebAPI(DatabasePage):
    __version__ = 1
    def __init__(self, config = {}):
        """
        __WebAPI__
        To create an auto-documenting API, make a subclass of the WebAPI class,
        add in methods that represent the api calls (no need to expose them),
        document them clearly in the methods doc strings and then add to the
        self.methods dictionary as follows:

        self.methods = {'method1':{'args':[],'call':self.method1},
                        'method2':{'args':['id'], 'call': self.method2}}

        WebAPI is a subclass of the DatabasePage to make a database connection
        available to the class in a standard way.

        binds = {'id': 123}
        sql = "select * from table where id = :id"
        result = self.dbi.processData(sql, binds)
        return self.formatDict(result)

        This can be used in the API methods. You may want to have a read-only
        API and a write-only API (with appropriately permissions on the DB
        connections) to minimise risk of SQL injection attacks (always use binds
        variables!).
        """
        DatabasePage.__init__(self, config)
        self.methods = {'status':{'args':[],
                                 'call':self.status,
                                 'version': 1}}

    def status(self):
        """
        Return the application name and it's version. Suitable for a ping.
        """
        return {'application': self.config.application,
                'version': self.__version__}

    @expose
    def stats(self):
        """
        Return CherryPy stats dict about underlying service activities
        """
        return cpstats.StatsPage().data()

    @expose
    def metrics(self):
        """
        Return CherryPy stats following the prometheus metrics structure
        """
        name = self.__class__.__name__
        metrics = promMetrics(cpstats.StatsPage().data(), name)
        return encodeUnicodeToBytes(metrics)

    @expose
    def index(self, **kwargs):
        """
        Return the auto-generated documentation for the API
        """
        return self.templatepage('API', methods = self.methods,
                                 webapp = self.config.application)

    @expose
    def default(self, *args, **kwargs):
        """
        Return the autogenerated documentation for the API (by calling index())
        TODO: decide whether to use args to call a method...
        """
        return self.index()

    #TODO: Validate input for json, das and xml
    @exposejson
    def json(self, *args, **kwargs):
        """
        The json output. args is assumed to be length 1 and contain the method
        name, kwargs are passed to the method
        """
        if len(args) > 0:
#            dict = self.runMethod(args[0], kwargs)
#            return dict
            res = self.runMethod(args[0], kwargs)
            if  isinstance(res, str):
                res = eval(res)
            return res
        else:
            raise HTTPRedirect("doc")

    def dasOutput(self, method, kwargs):
        """
        Make the API call and build a DAS compatible dictionary.
        """
        idict = {}
        try:
            idict['request_version'] = self.methods[method]['version']
        except:
            idict['request_version'] = 0
        idict['request_call'] = method
        idict[method] = self.runMethod(method, kwargs)
        return dict

    @exposedasjson
    def das(self, *args, **kwargs):
        """
        The das compatible json output. args is assumed to be length 1 and
        contain the method name, kwargs are passed to the method
        """
        try:
            return self.dasOutput(args[0], kwargs)
        except:
            raise HTTPRedirect("doc") from None

    @exposedasxml
    def dasxml(self, *args, **kwargs):
        """
        The das compatible xml output. args is assumed to be length 1 and
        contain the method name, kwargs are passed to the method
        """
        try:
            return self.dasOutput(args[0], kwargs)
        except:
            raise HTTPRedirect("doc")

    @exposedasplist
    def plist(self, *args, **kwargs):
        """
        The das compatible xml output. args is assumed to be length 1 and
        contain the method name, kwargs are passed to the method
        """
        try:
            res = self.runMethod(args[0], kwargs)
            if  isinstance(res, str):
                res = eval(res)
            return res
        except:
            raise HTTPRedirect("doc")

    @exposexml
    def xml(self, *args, **kwargs):
        """
        The xml output. args is assumed to be length 1 and contain the method
        name, kwargs are passed to the method
        """
        if len(args) > 0:
            res = self.runMethod(args[0], kwargs)
            if  isinstance(res, str):
                res = eval(res)
            return res
        else:
            raise HTTPRedirect("doc")

    def runMethod(self, method, userInput):
        """
        Run the specified method with the provided input, return a dict
        containing the result of the call or an exception wrapped in a dict.
        """
        idict = {}
        try:
            assert method in self.methods, "Unknown method called"
            if len(input):
                dict = self.methods[method]['call'](**input)
            else:
                idict = self.methods[method]['call']()
        except Exception as e:
            error = e.__str__()
            self.debug(error)
            self.debug(traceback.print_exc())
#            self.debug("%s:%s" % (sys.exc_type, sys.exc_value))
            idict = {'Exception':{'Exception_thrown_in': method,
                       'Exception_type': '%s' % sys.exc_info()[0],
                       'Exception_detail':error,
                       'Exception_arguments': input,
                       'Exception_dict':dict}}
        return idict



#!/usr/bin/env python
"""
_DeleteJobs_

Oracle implementation for creating a deleting a job
"""


from WMCore.BossAir.MySQL.DeleteJobs import DeleteJobs as MySQLDeleteJobs

class DeleteJobs(MySQLDeleteJobs):
    """
    _DeleteJobs_

    Delete jobs from bl_runjob
    """



#!/usr/bin/env python
"""
_BossAirPlugin_

Base class for BossAir plugins
"""


from WMCore.BossAir.Plugins.BasePlugin import BasePlugin



class TestPlugin(BasePlugin):
    """
    Test implementation of BasePlugin

    Does nothing
    """
    @staticmethod
    def stateMap():
        """
        For a given name, return a global state


        """

        stateDict = {'New': 'Pending',
                     'Gone': 'Error',
                     'Dead': 'Error'}

        # This call is optional but needs to for testing
        BasePlugin.verifyState(stateDict)
        return stateDict

    def __init__(self, config):

        BasePlugin.__init__(self, config)

        self.states = ['New', 'Dead', 'Gone']


    def submit(self, jobs, info = None):
        """
        Check and see if we have jobs

        """

        for job in jobs:
            #print job
            pass


        return jobs, []



    def track(self, jobs, info = None):
        """
        Label the jobs as done

        """

        return [], [], jobs



#!/usr/bin/env python
"""
_JobStatusByLocation_

Oracle implementation for loading a job by scheduler status
"""


from WMCore.BossAir.MySQL.JobStatusByLocation \
     import JobStatusByLocation as MySQLJobStatusByLocation

class JobStatusByLocation(MySQLJobStatusByLocation):
    pass



#!/usr/bin/env python
"""
File       : SiteListPoller
Author     : Valentin Kuznetsov <vkuznet AT gmail dot com>
Description: module to update of site lists within a WMAgent
"""

# system modules
import logging
import threading

# WMCore modules
from Utils.Timers import timeFunction
from WMCore.DAOFactory import DAOFactory
from WMCore.Services.WorkQueue.WorkQueue import WorkQueue
from WMCore.Services.WMStatsServer.WMStatsServer import WMStatsServer
from WMCore.WMSpec.WMWorkload import WMWorkloadHelper
from WMCore.WorkerThreads.BaseWorkerThread import BaseWorkerThread


class SiteListPoller(BaseWorkerThread):
    def __init__(self, config):
        """
        Initialize SiteListPoller object
        :param config: a Configuration object with the component configuration
        """
        BaseWorkerThread.__init__(self)
        myThread = threading.currentThread()
        self.logger = myThread.logger

        # get wmstats parameters
        self.wmstatsUrl = getattr(config.WorkflowUpdater, "wmstatsUrl")
        self.wmstatsSrv = WMStatsServer(self.wmstatsUrl, logger=self.logger)
        self.states = getattr(config.WorkflowUpdater, "states", ['running-open', 'acquired'])

        # provide access to WMBS in local WMAgent
        self.daoFactory = DAOFactory(package="WMCore.WMBS",
                                     logger=myThread.logger,
                                     dbinterface=myThread.dbi)
        # DB function to retrieve active workflows
        self.listActiveWflows = self.daoFactory(classname="Workflow.GetUnfinishedWorkflows")

        # local WorkQueue service
        self.localCouchUrl = config.WorkQueueManager.couchurl
        self.localWQ = WorkQueue(self.localCouchUrl,
                                 config.WorkQueueManager.dbname)

    def getActiveWorkflows(self):
        """
        Provide list of active requests within WMAgent
        :return: dict of workflows names vs pickle files
        """
        # get list of active workflows in WMAgent
        wflowSpecs = self.listActiveWflows.execute()

        # construct dictionary of workflow names and their pickle files
        wmaSpecs = {}
        for wflowSpec in wflowSpecs:
            name = wflowSpec['name']  # this is the name of workflow
            pklFileName = wflowSpec['spec']  # the "spec" in WMBS table (wmbs_workflow.spec) is pkl file name
            wmaSpecs[name] = pklFileName
        return wmaSpecs

    def wmstatsDict(self, requests):
        """
        Return list of requests specs from WMStats for provided list of request names
        :param requests: list of workflow requests names
        :return: dict of workflow records obtained from wmstats server:
        {"wflow": {"SiteWhitelist":[], "SiteBlacklist": []}, ...}
        """
        # get list of workflows from wmstats
        outputMask = ['SiteWhiteList', 'SiteBlackList']
        wdict = {}
        for state in self.states:
            inputConditions = {"RequestStatus": state}
            self.logger.info("Fetch site info from WMStats for condition: %s and mask %s", inputConditions, outputMask)
            data = self.wmstatsSrv.getFilteredActiveData(inputConditions, outputMask)
            for rdict in data:
                # rdict here has the following structure:
                # {"RequestName": "bla", "SiteWhitelist":[], "SiteBlacklist": []}
                wflow = rdict.pop('RequestName')
                # check that our workflow is in our requests list
                if wflow in requests:
                    wdict[wflow] = rdict
        return wdict

    @timeFunction
    def algorithm(self, parameters=None):
        """
        Perform the following logic:
        - obtain list of current active workflows from the agent
        - requests their specs from upstream wmstats server
        - update site lists of active workflows
        - push new specs to the agent local WorkQueue and update pickle spec file

        :return: none
        """
        # get list of active workflows from the agent, the returned dict
        # is composed by workflow names and associated pickle file (data comes from WMBS)
        wmaSpecs = self.getActiveWorkflows()
        wflows = wmaSpecs.keys()

        # obtain workflow records from wmstats server
        wdict = self.wmstatsDict(wflows)

        # iterate over the list of active workflows which is smaller than list from wmstats
        for wflow in wflows:
            if wflow not in wdict.keys():
                continue
            siteWhiteList = wdict[wflow]['SiteWhitelist']
            siteBlackList = wdict[wflow]['SiteBlacklist']

            # get the name of pkl file from wma spec
            pklFileName = wmaSpecs[wflow]

            # create wrapper helper and load pickle file
            wHelper = WMWorkloadHelper()
            wHelper.load(pklFileName)

            # extract from pickle spec both white and black site lists and compare them
            # to one we received from upstream service (ReqMgr2)
            wmaWhiteList = wHelper.getSiteWhitelist()
            wmaBlackList = wHelper.getSiteBlacklist()
            if set(wmaWhiteList) != set(siteWhiteList) or set(wmaBlackList) != set(siteBlackList):
                self.logger.info("Updating %s:", wflow)
                self.logger.info("  siteWhiteList %s => %s", wmaWhiteList, siteWhiteList)
                self.logger.info("  siteBlackList %s => %s", wmaBlackList, siteBlackList)
                try:
                    # update local WorkQueue first
                    params = {'SiteWhitelist': siteWhiteList, 'SiteBlacklist': siteBlackList}
                    self.localWQ.updateElementsByWorkflow(wHelper, params, status=['Available'])
                    self.logger.info("successfully updated workqueue elements for workflow %s", wflow)
                except Exception as ex:
                    logging.exception("Unexpected exception while updating elements in local workqueue Details:\n%s", str(ex))
                    continue

                # update workload only if we updated local WorkQueue
                # update site white/black lists together
                if set(wmaWhiteList) != set(siteWhiteList):
                    self.logger.info("updating site white list for workflow %s", wflow)
                    wHelper.setWhitelist(siteWhiteList)
                if set(wmaBlackList) != set(siteBlackList):
                    self.logger.info("updating site black list for workflow %s", wflow)
                    wHelper.setBlacklist(siteBlackList)

                try:
                    # persist the spec in local CouchDB
                    self.logger.info("Updating %s with new site lists within pkl file %s", wflow, pklFileName)
                    # save back pickle file
                    wHelper.save(pklFileName)
                except Exception as ex:
                    logging.exception("Caught unexpected exception in SiteListPoller. Details:\n%s", str(ex))
                    continue



#!/usr/bin/env python
"""
_AddToFilesByIDs_

Oracle implementation of AddFileToFilesetByIDs
"""

from WMCore.WMBS.MySQL.Files.AddToFilesetByIDs import AddToFilesetByIDs as AddFileToFilesetByIDsMySQL

class AddToFilesetByIDs(AddFileToFilesetByIDsMySQL):
    pass



"""
__UpdatePriority__

Oracle implementation of UpdatePriority

Created on Apr 16, 2013

@author: dballest
"""

from WMCore.WMBS.MySQL.Workflow.UpdatePriority import UpdatePriority as MySQLUpdatePriority

class UpdatePriority(MySQLUpdatePriority):
    pass



#!/usr/bin/env python
"""
_DBSBufferFile_

A simple object representing a file in DBSBuffer.
"""
from WMComponent.DBS3Buffer.DBSBufferDataset import DBSBufferDataset

from WMCore.DataStructs.File import File as WMFile
from WMCore.DataStructs.Run import Run
from WMCore.DAOFactory import DAOFactory
from WMCore.WMBS.WMBSBase import WMBSBase


class DBSBufferFile(WMBSBase, WMFile):

    def __init__(self, lfn=None, id=-1, size=None,
                 events=None, checksums=None, parents=None, locations=None,
                 status="NOTUPLOADED", inPhedex=0, workflowId=None, prep_id=None):

        checksums = checksums or {}
        WMBSBase.__init__(self)
        WMFile.__init__(self, lfn=lfn, size=size, events=events,
                        checksums=checksums, parents=parents, merged=True)
        self.setdefault("status", status)
        self.setdefault("in_phedex", inPhedex)
        self.setdefault("id", id)
        self.setdefault("workflowId", workflowId)

        # Parameters for the algorithm
        self.setdefault("appName", None)
        self.setdefault("appVer", None)
        self.setdefault("appFam", None)
        self.setdefault("psetHash", None)
        self.setdefault("configContent", None)
        self.setdefault("datasetPath", None)
        self.setdefault("processingVer", 0)
        self.setdefault("acquisitionEra", None)
        self.setdefault("validStatus", None)
        self.setdefault("globalTag", None)
        self.setdefault("datasetParent", None)
        self.setdefault("prep_id", None)

        if locations is None:
            self.setdefault("newlocations", set())
        else:
            self.setdefault("newlocations", self.makeset(locations))

        # The WMBS base class creates a DAO factory for WMBS, we'll need to
        # overwrite that so we can use the factory for DBSBuffer objects.
        self.daofactory = DAOFactory(package="WMComponent.DBS3Buffer",
                                     logger=self.logger,
                                     dbinterface=self.dbi)

        return

    def exists(self):
        """
        _exists_

        Determine whether or not a file with this LFN exists inside the
        database.  Return the file's ID if it exists, False otherwise.
        """
        action = self.daofactory(classname="DBSBufferFiles.Exists")
        return action.execute(lfn=self["lfn"], conn=self.getDBConn(),
                              transaction=self.existingTransaction())

    def getStatus(self):
        """
        _getStatus_

        Retrieve the status of the file.  This can be one of the following:
          UPLOADED
          NOTUPLOADED
        """
        return self["status"]

    def getLocations(self):
        """
        _getLocations_

        Retrieve a list of locations where this file is stored.
        """
        return list(self["locations"])

    def getRuns(self):
        """
        _getRuns_

        Retrieve a list of WMCore.DataStructs.Run objects that represent which
        run/lumi sections this file contains.
        """
        return list(self["runs"])

    def load(self, parentage=0):
        """
        _load_

        The the file and all it's metadata from the database.  Either the LFN
        or the file's ID must be specified before this is called.
        """
        existingTransaction = self.beginTransaction()

        if self["id"] != -1:
            action = self.daofactory(classname="DBSBufferFiles.GetByID")
            result = action.execute(self["id"], conn=self.getDBConn(),
                                    transaction=self.existingTransaction())
        else:
            action = self.daofactory(classname="DBSBufferFiles.GetByLFN")
            result = action.execute(self["lfn"], conn=self.getDBConn(),
                                    transaction=self.existingTransaction())

        self.update(result)

        action = self.daofactory(classname='DBSBufferFiles.GetChecksum')
        result = action.execute(fileid=self['id'], conn=self.getDBConn(),
                                transaction=self.existingTransaction())
        self["checksums"] = result

        action = self.daofactory(classname="DBSBufferFiles.GetRunLumiFile")
        runs = action.execute(self["lfn"], conn=self.getDBConn(),
                              transaction=self.existingTransaction())
        for r in runs:
            self.addRun(run=Run(r, *runs[r]))

        action = self.daofactory(classname="DBSBufferFiles.GetLocation")
        self["locations"] = action.execute(self["lfn"], conn=self.getDBConn(),
                                           transaction=self.existingTransaction())

        self["newlocations"].clear()
        self["parents"].clear()

        if parentage > 0:
            action = self.daofactory(classname="DBSBufferFiles.GetParents")
            lfns = action.execute(self["lfn"], conn=self.getDBConn(),
                                  transaction=self.existingTransaction())
            for lfn in lfns:
                parentFile = DBSBufferFile(lfn=lfn)
                parentFile.load(parentage=parentage - 1)
                self["parents"].add(parentFile)

        self.commitTransaction(existingTransaction)
        return

    def insertDatasetAlgo(self):
        """
        _insertDatasetAlgo_

        Insert the dataset and algorithm for this file into the DBS Buffer.
        """
        newAlgoAction = self.daofactory(classname="NewAlgo")
        assocAction = self.daofactory(classname="AlgoDatasetAssoc")

        existingTransaction = self.beginTransaction()

        newAlgoAction.execute(appName=self["appName"], appVer=self["appVer"],
                              appFam=self["appFam"], psetHash=self["psetHash"],
                              configContent=self["configContent"],
                              conn=self.getDBConn(),
                              transaction=True)

        dbsbufferDataset = DBSBufferDataset(self["datasetPath"],
                                            processingVer=self['processingVer'],
                                            acquisitionEra=self['acquisitionEra'],
                                            validStatus=self['validStatus'],
                                            globalTag=self.get('globalTag', None),
                                            parent=self['datasetParent'],
                                            prep_id=self['prep_id'])

        if dbsbufferDataset.exists():
            dbsbufferDataset.updateDataset()
        else:
            dbsbufferDataset.create()

        assocID = assocAction.execute(appName=self["appName"], appVer=self["appVer"],
                                      appFam=self["appFam"], psetHash=self["psetHash"],
                                      datasetPath=self["datasetPath"],
                                      conn=self.getDBConn(),
                                      transaction=True)

        self.commitTransaction(existingTransaction)
        return assocID

    def create(self):
        """
        _create_

        Insert this file and all it's metadata into the DBS Buffer.
        """

        if self.exists() != False:
            self.load()
            return

        existingTransaction = self.beginTransaction()
        assocID = self.insertDatasetAlgo()

        addAction = self.daofactory(classname="DBSBufferFiles.Add")
        addAction.execute(files=self["lfn"], size=self["size"],
                          events=self["events"],
                          datasetAlgo=assocID, status=self["status"],
                          workflowID=self["workflowId"],
                          conn=self.getDBConn(),
                          transaction=self.existingTransaction())

        if len(self["runs"]) > 0:
            lumiAction = self.daofactory(classname="DBSBufferFiles.AddRunLumi")
            lumiAction.execute(file=self["lfn"], runs=self["runs"],
                               conn=self.getDBConn(),
                               transaction=self.existingTransaction())

        self["id"] = self.exists()
        self.updateLocations()
        self.commitTransaction(existingTransaction)

        for checksumType in self["checksums"]:
            self.setCksum(cksum=self["checksums"][checksumType],
                          cktype=checksumType)
        return

    def delete(self):
        """
        _delete_

        Remove a file from the DSBuffer database.
        """
        action = self.daofactory(classname="DBSBufferFiles.Delete")
        action.execute(file=self["lfn"], conn=self.getDBConn(),
                       transaction=self.existingTransaction())
        return

    def addChildren(self, lfns):
        """
        _addChildren_

        Set one or more lfns as the child of this file.
        """
        if not isinstance(lfns, list):
            lfns = [lfns]

        existingTransaction = self.beginTransaction()

        if not self["id"] > 0:
            raise Exception("Parent file doesn't have an id %s" % self["lfn"])

        action = self.daofactory(classname="DBSBufferFiles.HeritageLFNChild")
        action.execute(childLFNs=lfns, parentID=self["id"],
                       conn=self.getDBConn(),
                       transaction=self.existingTransaction())

        self.commitTransaction(existingTransaction)
        return

    def addParents(self, parentLFNs):
        """
        _addParents_

        Associate this file with it's parents.  If the parents do not exist in
        the buffer then bogus place holder files will be created so that the
        parentage information can be tracked and correctly inserted into DBS.
        """
        newAlgoAction = self.daofactory(classname="NewAlgo")
        newDatasetAction = self.daofactory(classname="NewDataset")
        assocAction = self.daofactory(classname="AlgoDatasetAssoc")
        existsAction = self.daofactory(classname="DBSBufferFiles.Exists")

        uploadFactory = DAOFactory(package="WMComponent.DBS3Buffer",
                                   logger=self.logger,
                                   dbinterface=self.dbi)
        setDatasetAlgoAction = uploadFactory(classname="SetDatasetAlgo")

        existingTransaction = self.beginTransaction()

        toBeCreated = []
        for parentLFN in parentLFNs:
            self["parents"].add(DBSBufferFile(lfn=parentLFN))
            if not existsAction.execute(lfn=parentLFN,
                                        conn=self.getDBConn(),
                                        transaction=True):
                toBeCreated.append(parentLFN)

        if len(toBeCreated) > 0:
            newAlgoAction.execute(appName="cmsRun", appVer="UNKNOWN",
                                  appFam="UNKNOWN", psetHash="NOT_SET",
                                  configContent="NOT_SET",
                                  conn=self.getDBConn(),
                                  transaction=True)

            newDatasetAction.execute(datasetPath="bogus",
                                     conn=self.getDBConn(),
                                     transaction=True)

            assocID = assocAction.execute(appName="cmsRun", appVer="UNKNOWN",
                                          appFam="UNKNOWN", psetHash="NOT_SET",
                                          datasetPath="bogus",
                                          conn=self.getDBConn(),
                                          transaction=True)

            setDatasetAlgoAction.execute(datasetAlgo=assocID, inDBS=1,
                                         conn=self.getDBConn(),
                                         transaction=True)

            addInputFiles = self.daofactory(classname="DBSBufferFiles.AddIgnore")
            addInputFiles.execute(lfns=toBeCreated, datasetAlgo=assocID,
                                  status="GLOBAL",
                                  inPhEDEx=1,
                                  conn=self.getDBConn(),
                                  transaction=True)

        action = self.daofactory(classname="DBSBufferFiles.HeritageLFNParent")
        action.execute(parentLFNs=parentLFNs, childLFN=self["lfn"],
                       conn=self.getDBConn(),
                       transaction=self.existingTransaction())
        self.commitTransaction(existingTransaction)
        return

    def updateLocations(self):
        """
        _updateLocations_

        Write any new locations to the database.  After any new locations are
        written to the database all locations will be reloaded from the
        database.
        """
        if len(self["newlocations"]) == 0:
            return

        insertAction = self.daofactory(classname="DBSBufferFiles.AddLocation")
        addAction = self.daofactory(classname="DBSBufferFiles.SetLocationByLFN")

        existingTransaction = self.beginTransaction()

        insertAction.execute(siteName=self['newlocations'],
                             conn=self.getDBConn(),
                             transaction=self.existingTransaction())

        binds = []
        for location in self["newlocations"]:
            binds.append({"lfn": self["lfn"],
                          "pnn": location})

        addAction.execute(binds=binds,
                          conn=self.getDBConn(),
                          transaction=self.existingTransaction())

        self["locations"].update(self["newlocations"])
        self["newlocations"].clear()
        self.commitTransaction(existingTransaction)
        return

    def setLocation(self, pnn, immediateSave=True):
        """
        _setLocation_

        Sets the location of a file. If immediateSave is True, commit change to
        the DB immediately, otherwise queue for addition when save() is called.
        """
        if isinstance(pnn, str):
            self["newlocations"].add(pnn)
            self["locations"].add(pnn)
        else:
            self["newlocations"].update(pnn)
            self["locations"].update(pnn)

        if immediateSave:
            self.updateLocations()

        return

    def setAlgorithm(self, appName=None, appVer=None, appFam=None,
                     psetHash=None, configContent=None):
        """
        _setAlgorithm_

        Set the DBS algorithm for this file.
        """
        self["appName"] = appName
        self["appVer"] = appVer
        self["appFam"] = appFam
        self["psetHash"] = psetHash
        self["configContent"] = configContent
        return

    def setProcessingVer(self, ver):
        """
        _setProcessingVer_

        Set the era
        """
        procVer = ver or 0
        try:
            procVer = int(procVer)
        except TypeError:
            msg = "Provided ver=%s of type %s cannot be converted to int" \
                    % (ver, type(ver))
            raise TypeError(msg) from None
        self['processingVer'] = procVer
        return

    def setAcquisitionEra(self, era):
        """
        _setAcquisitionEra_

        Set the era
        """

        self['acquisitionEra'] = era
        return

    def setDatasetPath(self, datasetPath):
        """
        _setDatasetPath_

        Set the dataset path for this file.
        """
        self["datasetPath"] = datasetPath
        return

    def setValidStatus(self, validStatus):
        """
        _setValidStatus_

        Set the valid status for the dataset to be migrated into DBS
        """

        self['validStatus'] = validStatus
        return

    def setGlobalTag(self, globalTag):
        """
        _setGlobalTag_

        Set the global Tag
        """

        self['globalTag'] = globalTag
        return

    def getGlobalTag(self):
        """
        _getGlobalTag_

        Get the global Tag
        """

        return self['globalTag']

    def setPrepID(self, prep_id):

        self['prep_id'] = prep_id
        return

    def getPrepID(self):

        return self['prep_id']

    def setDatasetParent(self, datasetParent):
        """
        _setDatasetParent_

        Set the dataset's parent path
        """

        self['datasetParent'] = datasetParent
        return

    def getDatasetParent(self):
        """
        _getDatasetParent_

        Return the datasetParent
        """

        return self['datasetParent']

    def getParentLFNs(self):
        """
        Get a flat list of parent LFNs
        """
        result = []
        parents = self['parents']
        while parents:
            result.extend(parents)
            temp = []
            for parent in parents:
                temp.extend(parent["parents"])
            parents = temp
        result.sort()  # ensure SecondaryInputFiles are in order
        return [x['lfn'] for x in result]

    def addRunSet(self, runSet):
        """
        add the set of runs.  This should be called after a file is created,
        unlike addRun which should be called before the file was created.
        runSet should be set of DataStruct.Run
        also there should be no duplicate entries in runSet.
        (May need to change in schema level not to allow duplicate record)
        """
        existingTransaction = self.beginTransaction()

        lumiAction = self.daofactory(classname="DBSBufferFiles.AddRunLumi")
        lumiAction.execute(file=self["lfn"], runs=runSet,
                           conn=self.getDBConn(),
                           transaction=self.existingTransaction())

        action = self.daofactory(classname="DBSBufferFiles.GetRunLumiFile")
        runs = action.execute(self["lfn"], conn=self.getDBConn(),
                              transaction=self.existingTransaction())

        self["runs"].clear()
        for r in runs:
            self.addRun(run=Run(r, *runs[r]))

        self.commitTransaction(existingTransaction)
        return

    def setBlock(self, blockName):
        """
        _setBlock_

        Associate this file with a block in DBS/PhEDEx.
        """
        existingTransaction = self.beginTransaction()

        blockAction = self.daofactory(classname="DBSBufferFiles.SetBlock")
        blockAction.execute(self["lfn"], blockName, conn=self.getDBConn(),
                            transaction=self.existingTransaction())

        self.commitTransaction(existingTransaction)
        return

    def setCksum(self, cksum, cktype):
        """
        _setCKSum_

        Set the Checksum
        """
        if self['id'] < 0:
            self.create()

        existingTransaction = self.beginTransaction()

        action = self.daofactory(classname="DBSBufferFiles.AddChecksum")
        action.execute(fileid=self['id'], cktype=cktype, cksum=cksum, \
                       conn=self.getDBConn(), transaction=existingTransaction)

        self.commitTransaction(existingTransaction)

        return



#!/usr/bin/env python
"""
_JobGroup_

Definition of JobGroup:
    Set of jobs running on same input file for same Workflow
    Set of jobs for a single subscription
    Required for certain job splitting Algo's (.g. event split to make complete
    lumi)
    Subscription:JobGroup == 1:N
    JobGroup:Jobs = 1:N
    JobGroup:InFile = 1:1
    JobGroup:MergedOutFile = N:1
    JobGroup at least one Lumi section

A JobGroup is a set of jobs and a Fileset that contains their output.

JobGroup knows the Subscription and passes the Workflow to Jobs in the group.

Jobs know their status (active, failed, complete) and know the files they run
on but don't know the group. They do know their subscription and corresponding
workflow. This means Jobs can update their state in the database without
talking to the group, and WMBS JobGroups can calculate status from the database
instead of the in memory objects.
"""

import logging
import threading

from WMCore.DataStructs.JobGroup import JobGroup as WMJobGroup
from WMCore.WMBS.WMBSBase import WMBSBase

from WMCore.WMBS.Fileset import Fileset
from WMCore.WMBS.Job import Job
from WMCore.WMBS.Subscription import Subscription

from WMCore.Services.UUIDLib import makeUUID

class JobGroup(WMBSBase, WMJobGroup):
    """
    A group (set) of Jobs
    """
    def __init__(self, subscription = None, jobs = None, id = -1, uid = None, location = None):
        WMBSBase.__init__(self)
        WMJobGroup.__init__(self, subscription=subscription, jobs = jobs)

        self.id = id
        self.lastUpdate = None
        self.uid = uid

        if location != None:
            self.setSite(location)

        return

    def create(self):
        """
        Add the new jobgroup to WMBS, create the output Fileset object
        """
        myThread = threading.currentThread()
        existingTransaction = self.beginTransaction()

        #overwrite base class self.output for WMBS fileset
        self.output = Fileset(name = makeUUID())
        self.output.create()

        if self.uid == None:
            self.uid = makeUUID()

        action = self.daofactory(classname = "JobGroup.New")
        action.execute(self.uid, self.subscription["id"],
                       self.output.id, conn = self.getDBConn(),
                       transaction = self.existingTransaction())

        self.id = self.exists()
        self.commitTransaction(existingTransaction)

        return

    def delete(self):
        """
        Remove a jobgroup from WMBS
        """
        deleteAction = self.daofactory(classname = "JobGroup.Delete")
        deleteAction.execute(id = self.id, conn = self.getDBConn(),
                             transaction = self.existingTransaction())

        return

    def exists(self):
        """
        Does a jobgroup exist with id if id is not provided, use the uid,
        return the id
        """
        if self.id != -1:
            action = self.daofactory(classname = "JobGroup.ExistsByID")
            result =  action.execute(id = self.id, conn = self.getDBConn(),
                                     transaction = self.existingTransaction())
        else:
            action = self.daofactory(classname = "JobGroup.Exists")
            result = action.execute(uid = self.uid, conn = self.getDBConn(),
                                    transaction = self.existingTransaction())

        return result

    def load(self):
        """
        _load_

        Load all meta data associated with the JobGroup.  This includes the
        JobGroup id, uid, last_update time, subscription id and output fileset
        id.  Either the JobGroup id or uid must be specified for this to work.
        """
        existingTransaction = self.beginTransaction()

        if self.id > 0:
            loadAction = self.daofactory(classname = "JobGroup.LoadFromID")
            result = loadAction.execute(self.id, conn = self.getDBConn(),
                                        transaction = self.existingTransaction())
        else:
            loadAction = self.daofactory(classname = "JobGroup.LoadFromUID")
            result = loadAction.execute(self.uid, conn = self.getDBConn(),
                                        transaction = self.existingTransaction())

        self.id = result["id"]
        self.uid = result["uid"]
        self.lastUpdate = result["last_update"]

        self.subscription = Subscription(id = result["subscription"])
        self.subscription.load()

        self.output = Fileset(id = result["output"])
        self.output.load()

        self.jobs = []
        self.commitTransaction(existingTransaction)
        return

    def loadData(self):
        """
        _loadData_

        Load all data that is associated with the jobgroup.  This includes
        loading all the subscription information, the output fileset
        information and all the jobs that are associated with the group.
        """
        existingTransaction = self.beginTransaction()

        if self.id < 0 or self.uid == None:
            self.load()

        self.subscription.loadData()
        self.output.loadData()

        loadAction = self.daofactory(classname = "JobGroup.LoadJobs")
        result = loadAction.execute(self.id, conn = self.getDBConn(),
                                    transaction = self.existingTransaction())

        self.jobs = []
        self.newjobs = []

        for jobID in result:
            newJob = Job(id = jobID["id"])
            newJob.loadData()
            self.add(newJob)

        WMJobGroup.commit(self)
        self.commitTransaction(existingTransaction)
        return

    def commit(self):
        """
        _commit_

        Write any new jobs to the database, creating them in the database if
        necessary.
        """
        existingTransaction = self.beginTransaction()

        if self.id == -1:
            self.create()

        for j in self.newjobs:
            j.create(group = self)

        WMJobGroup.commit(self)
        self.commitTransaction(existingTransaction)
        return


    def setSite(self, site_name = None):
        """
        Updates the jobGroup with a site_name from the wmbs_location table
        """
        if not self.exists():
            return

        action = self.daofactory(classname = "JobGroup.SetSite")
        result = action.execute(site_name = site_name, jobGroupID = self.id,
                                conn = self.getDBConn(), transaction = self.existingTransaction())

        return result


    def getSite(self):
        """
        Updates the jobGroup with a site_name from the wmbs_location table
        """
        if not self.exists():
            return

        action = self.daofactory(classname = "JobGroup.GetSite")
        result = action.execute(jobGroupID = self.id, conn = self.getDBConn(),
                                transaction = self.existingTransaction())

        return result

    def listJobIDs(self):
        """
        Returns a list of job IDs
        Useful for times when threading the loading of jobGroups, where running loadData can overload UUID
        """

        existingTransaction = self.beginTransaction()

        if self.id < 0 or self.uid == None:
            self.load()

        loadAction = self.daofactory(classname = "JobGroup.LoadJobs")
        result = loadAction.execute(self.id, conn = self.getDBConn(),
                                    transaction = self.existingTransaction())

        jobIDList = []

        for jobID in result:
            jobIDList.append(jobID["id"])

        self.commitTransaction(existingTransaction)
        return jobIDList


    def commitBulk(self):
        """
        Creates jobs in a group instead of singly, as is done in jobGroup.commit()
        """

        myThread = threading.currentThread()

        if self.id == -1:
            myThread.transaction.begin()
            #existingTransaction = self.beginTransaction()
            self.create()
            #self.commitTransaction(existingTransaction)
            myThread.transaction.commit()

        existingTransaction = self.beginTransaction()

        listOfJobs = []
        for job in self.newjobs:
            #First do all the header stuff
            if job["id"] != None:
                continue

            job["jobgroup"] = self.id

            if job["name"] == None:
                job["name"] = makeUUID()

            listOfJobs.append(job)

        bulkAction = self.daofactory(classname = "Jobs.New")
        result = bulkAction.execute(jobList = listOfJobs)

        #Use the results of the bulk commit to get the jobIDs
        fileDict = {}
        for job in listOfJobs:
            job['id'] = result[job['name']]
            fileDict[job['id']] = []
            for file in job['input_files']:
                fileDict[job['id']].append(file['id'])

        maskAction = self.daofactory(classname = "Masks.New")
        maskAction.execute(jobList = listOfJobs, conn = self.getDBConn(), \
                           transaction = self.existingTransaction())

        fileAction = self.daofactory(classname = "Jobs.AddFiles")
        fileAction.execute(jobDict = fileDict, conn = self.getDBConn(), \
                           transaction = self.existingTransaction())


        WMJobGroup.commit(self)
        self.commitTransaction(existingTransaction)

        return


    def getLocationsForJobs(self):
        """
        Gets a list of the locations that jobs can run at
        """
        if not self.exists():
            return

        action = self.daofactory(classname = "JobGroup.GetLocationsForJobs")
        result = action.execute(id = self.id, conn = self.getDBConn(),
                                transaction = self.existingTransaction())

        return result


    def __str__(self):
        """
        __str__

        Print out some information about the jobGroup
        as if jobGroup inherited from dict()
        """

        d = {'id': self.id, 'uid': self.uid, 'subscription': self.subscription,
             'output': self.output, 'jobs': self.jobs,
             'newjobs': self.newjobs}

        return str(d)
